"""Fllistview module."""
# -*- coding: utf-8 -*-

from PyQt6 import QtGui  # type: ignore
from pineboolib.core import decorators
from pineboolib import logging
from pineboolib.q3widgets import qlistview
from typing import Any, Optional, cast, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # noqa: F401 # pragma: no cover

LOGGER = logging.get_logger("FLListViewItem")


class FLListViewItem(QtGui.QStandardItem):
    """FLListView class."""

    _expandable: bool
    _key: str
    _open: bool
    _root: bool
    _index_child: int

    def __init__(self, parent: Optional["QtWidgets.QWidget"] = None) -> None:
        """Inicialize."""

        super().__init__()
        self._root = False
        self.setOpen(False)
        self.setExpandable(False)
        self._parent = None
        self.setKey("")
        self.setEditable(False)
        self._index_child = 0
        if parent:
            # Comprueba que tipo de parent es
            if isinstance(parent, qlistview.QListView):
                # self._root = True
                parent.model().setItem(0, 0, self)
            else:
                if isinstance(parent, self):  # type: ignore [arg-type] # noqa: F821
                    # print("Añadiendo nueva linea a", parent.text(0))
                    cast(FLListViewItem, parent).appendRow(self)

        # if parent:
        #    self._parent = parent
        #    self._row = self._parent.model().rowCount()
        #    if self._parent.model().item(0,0) is not None:
        #        self._parent.model().item(0,0).setChild(self._row,0, self)
        #        self._parent.model().item(0,0)._rowcount += 1
        #    else:
        #        self._parent.model().setItem(self._row,0,self)

        #    self._rows = self._parent.model().item(0,0)._rowcount - 1

    def firstChild(self) -> Any:
        """Return first child."""

        self._index_child = 0
        item = self.child(self._index_child)
        return item

    def nextSibling(self) -> Any:
        """Return next child."""

        self._index_child += 1
        item = self.child(self._index_child)
        return item

    def isExpandable(self) -> bool:
        """Return if is expandable."""

        return self._expandable
        # return True if self.child(0) is not None or not self.parent() else False

    def setText(self, *args) -> None:
        """Set text."""

        # print("Seteando", args, self.parent())
        # LOGGER.warning("Seteo texto %s" , args, stack_info = True )
        col = 0
        if len(args) == 1:
            value = args[0]
        else:
            col = args[0]
            value = str(args[1])

        if col == 0:
            # if self._root:
            # print("Inicializando con %s a %s" % ( value, self.parent()))
            super().setText(value)
        else:
            item = self.parent().child(self.row(), col)  # type: ignore [union-attr]
            if item is None:
                item = FLListViewItem()
                self.parent().setChild(self.row(), col, item)  # type: ignore [union-attr]

            item.setText(value)

    def text(self, col: int = 0) -> str:
        """Return text from a column."""

        ret = ""
        if col == 0:
            ret = super().text()

        return str(ret)

    @decorators.not_implemented_warn
    def setPixmap(self, *args):
        """Set pixmap."""
        pass

    def setExpandable(self, val: bool) -> None:
        """Set expandable."""

        self._expandable = val

    def setKey(self, key: str) -> None:
        """Set key."""
        self._key = key

    def key(self) -> str:
        """Return key."""

        if self.parent() and self.column() > 0:
            return (
                self.parent().child(self.row(), 0).key()  # type: ignore [attr-defined, union-attr] # noqa: F821
            )
        return self._key

    def setOpen(self, open: bool) -> None:
        """Set Open."""
        self._open = open



# -*- coding: utf-8 -*-
"""Check application dependencies."""
import sys
import traceback
from typing import Dict, Tuple

from pineboolib.core.utils import logging
from pineboolib.core.utils.version import VersionNumber

ModuleName = str
ModuleVersion = str
SuggestedPackageNameForInstall = str
ErrorString = str

DependencyCheck = Dict["ModuleName", "SuggestedPackageNameForInstall"]
DependencyError = Dict[Tuple["ModuleName", "SuggestedPackageNameForInstall"], "ErrorString"]

LOGGER = logging.get_logger(__name__)
DEPENDENCIES_CHECKED: Dict["ModuleName", "ModuleVersion"] = {}

MINIMUM_VERSION = {
    "Python": "3.6",
    "ply": "3.9",
    "Pillow": "5.1.0",
    "fpdf": "1.7.3",
    "PyQt6": "5.11",
}
PYTHON_INCLUDED_BATTERIES = {"Python", "sqlite3"}


def get_dependency_errors(dict_: "DependencyCheck") -> "DependencyError":
    """
    Check if a package is installed and return the result.

    @param dict_. Dict with the name of the agency and the module to be checked.
    @param exit . Exit if dependence fails.
    """

    global DEPENDENCIES_CHECKED
    from importlib import import_module

    error: DependencyError = {}
    for key, suggested_pkg in dict_.items():
        if key in DEPENDENCIES_CHECKED:
            continue
        version = None
        if key in PYTHON_INCLUDED_BATTERIES:
            version = sys.version[: sys.version.find("(")]
        else:
            try:
                mod_ = import_module(key)
                version = getattr(mod_, "__version__", None)
                if key == "PyQt6.QtCore":
                    version = getattr(mod_, "QT_VERSION_STR", None)
                del mod_
            except ImportError:
                error[(key, suggested_pkg)] = traceback.format_exc()
                continue

        if key == "odf":
            from odf import namespaces  # type: ignore

            version = namespaces.__version__

        elif key == "barcode":
            import barcode  # type: ignore

            version = barcode.version
        elif key == "pyodbc":
            import pyodbc  # type: ignore

            version = pyodbc.version

        if not isinstance(version, str):
            error[(key, suggested_pkg)] = (
                "Error: version detected should be string, but found %r" % version
            )
            continue

        if key in MINIMUM_VERSION:
            VersionNumber.check(key, version, MINIMUM_VERSION[key])

        DEPENDENCIES_CHECKED[key] = version
        LOGGER.debug("Dependency checked %s: %s", key, version)
    return error


def check_dependencies_cli(dict_: "DependencyCheck") -> bool:
    """
    Check if a package is installed and return the result.

    @param dict_. Dict with the name of the agency and the module to be checked.
    @param exit . Exit if dependence fails.
    """
    dep_error: DependencyError = get_dependency_errors(dict_)
    if not dep_error:
        return True

    LOGGER.warning("Unmet dependences:")
    for (dep, suggestedpkg), errormsg in dep_error.items():
        LOGGER.warning("... Error trying to import module %s:\n%s", dep, errormsg)
        LOGGER.warning("... Install package %s for %s", suggestedpkg, dep)

    return False



# -*- coding: utf-8 -*-
r"""
Base class for Access Controls, also called Access Control Rules.

An access control rule applies to a user and a high-level object or container
(main windows, tables, etc.), which in turn will contain other objects (actions, fields, etc.).
The rule is defined by the following information as its header, which identifies it
uniquely:

\\ code

         kind           ; first name         ; user           ; permission
-------------------------------------------------- ------------------------------------------------
PNAccessControl :: type; PNAccessControl :: name; PNAccessControl :: user; PNAccessControl :: perm

\\ endcode

The type will be that of the high level object, the name will be that of the object, the user will correspond to the
name of the user in the database to which the rule applies and permission will be an identifier
of text that defines the type of permission attributed to the object for the given user. This permission
It is general or global and will be applied by default to all child objects that belong to the object
high level.

At the same time a rule may have a list of Access Control Objects (called ACOs,
Access Control Objects), to which you want to apply a permission other than the general one. The ACOs will be
child objects or belonging to the high level object. Internally the list of ACOs is composed of
tuples of two elements; 'object name' and 'permission', the object name will be the one assigned
within the hierarchy of objects belonging to the high level object and permission will be the permission for that
object and that will overwrite the general permit.

The values ​​of the rule can be set from a DOM node of a DOM / XML document, using
PNAccessControl :: set. Reciprocally you can obtain a DOM node with the content of the rule,
to insert into a DOM / XML document, using PNAccessControl :: get. The general XML structure of the DOM node
which represents an access control rule is as follows:

\\ code

<[mainwindow, table, etc ..] perm = 'XXX'>
  <name> XXX </name>
  <user> XXX </user>
  <aco perm = 'XXX'> XXX </aco>
  ....
  <aco perm = 'XXX'> XXX </aco>
 </ [mainwindow, table, etc ..]>

\\ endcode

For convenience, the PNAccessControl :: setAcos method is also provided, which allows you to establish the list of
ACOs of a rule directly from a list of text strings.

This class is not intended to be used directly, but as a basis for derived classes that are
specifically responsible for processing high-level objects. An example would be PNAccessControlMainWindow,
which is responsible for access control for high-level objects of type 'mainwindow', ie main windows,
as the module selector, or each of the main windows of the modules.

@author InfoSiAL S.L.
"""


from typing import List, Dict, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtXml, QtWidgets  # pragma: no cover
    from pineboolib.application.metadata import pntablemetadata  # pragma: no cover


class PNAccessControl(object):
    """PNAccessControl Class."""

    """
    Almacena el nombre del objeto de alto nivel.
    """

    _name: str
    """
    Almacena el nombre del usuario de la base de datos.
    """
    _user: str
    """
    Almacena el permiso general de la regla de control de acceso.
    """
    _perm: str

    """
    Diccionario de permisos específicos de los ACOs (Access Control Objects)
    hijos o pertenecientes al objeto de alto nivel. El diccionario almacena la
    correspondencia entre el nombre del ACO (utilizado como clave de búsqueda)
    y el permiso a aplicar.
    """
    _acos_perms: Dict[str, str]

    def __init__(self) -> None:
        """Initialize."""
        self._name = ""
        self._user = ""
        self._perm = ""
        self._acos_perms = {}

    def __del__(self) -> None:
        """Remove values ​​when closed."""

        self._acos_perms.clear()
        del self._acos_perms

    def name(self) -> str:
        """
        Get the name of the high level object.

        @return Text string with the name of the object.
        """

        return self._name

    def user(self) -> str:
        """
        Get the name of the database user.

        @return Text string with the user's name (login).
        """

        return self._user

    def perm(self) -> str:
        """
        Get general permission.

        @return Text string that identifies the permission to apply.
        """

        return self._perm

    def setName(self, name: str) -> None:
        """
        Set the name of the high level object.

        @param n Object name.
        """

        self._name = name

    def setUser(self, user: str) -> None:
        """
        Set the name of the database user.

        @param u Name (login) of the user.
        """

        self._user = user

    def setPerm(self, perm: str) -> None:
        """
        Set the general permission.

        @param p Text string with the permission identifier.
        """

        self._perm = perm

    def clear(self) -> None:
        """
        Clean the rule by emptying it and freeing all resources.
        """

        self._name = ""
        self._user = ""
        self._perm = ""

        self._acos_perms.clear()

    def type(self) -> str:
        """
        Return a text constant that identifies the type.

        This function must be reimplemented in the derived classes that are
        responsible for processing a specific object type and return the relevant identifier

        @return A text string that identifies the general object type of the rule, eg "table".
        """

        return ""

    def set(self, element: Optional["QtXml.QDomElement"]) -> None:
        """
        Define the access control rule from the information of a DOM node of a given DOM / XML document.

        @param e Element corresponding to the DOM node that will be used to define the rule.
        """

        if element is None:
            return

        self._acos_perms.clear()

        self._perm = element.attribute("perm")
        node = element.firstChild()

        while not node.isNull():
            node_elem = node.toElement()
            if not node_elem.isNull():
                tag_name = node_elem.tagName()
                if tag_name == "name":
                    self._name = node_elem.text()

                elif tag_name == "user":
                    self._user = node_elem.text()

                elif tag_name == "aco":
                    self._acos_perms[node_elem.text()] = node_elem.attribute("perm")

            node = node.nextSibling()

    def get(self, document: Optional["QtXml.QDomDocument"]) -> None:
        """
        From the content of the access control rule create a DOM node.

        This will be inserted as child of the first node of a DOM / XML document.

        @param d DOM / XML document where the node built from the access control rule will be inserted.
        """

        if document is None:
            return  # type: ignore [unreachable] # noqa: F821

        type_ = self.type()
        if not type_:
            return

        root = document.firstChild().toElement()
        element = document.createElement(type_)
        element.setAttribute("perm", self._perm)
        root.appendChild(element)

        name_element = document.createElement("name")
        name_element.appendChild(document.createTextNode(self._name))
        element.appendChild(name_element)

        user_element = document.createElement("user")
        user_element.appendChild(document.createTextNode(self._user))
        element.appendChild(user_element)

        for key in self._acos_perms.keys():
            aco_element = document.createElement("aco")
            aco_element.setAttribute("perm", self._acos_perms[key])
            aco_element.appendChild(document.createTextNode(key))
            element.appendChild(aco_element)

    def setAcos(self, acos: List[str]) -> None:
        """
        Set the list of Acos from a list of text strings.

        This list of texts must have in their order components the names of the objects, and in the
        odd order components the permission to apply to that object, eg: "pbOpen", "r-", "lblText", "-", "tbErase", "rw", ...

        @param acos List of text strings with objects and permissions.
        """

        self._acos_perms.clear()
        i = 0
        while i < len(acos):
            self._acos_perms[acos[i]] = acos[i + 1]
            i += 2

    def getAcos(self) -> List[str]:
        """
        Get a list of text strings corresponding to the list of ACOs set.

        The format of this list is the same as described in PNAccessControl :: setAcos
        eg 'pbOpen', 'r-', 'lblText', '--', 'tbErase', 'rw', ...

        @return List of text strings with objects and permissions.
        """

        acos = []

        for key, value in self._acos_perms.items():
            acos.append(key)
            acos.append(value)

        return acos

    def processObject(
        self,
        obj: Union["QtWidgets.QMainWindow", "QtWidgets.QWidget", "pntablemetadata.PNTableMetaData"],
    ) -> None:
        """Process object overload function."""

        return None



"""Modules module."""


from pineboolib.core.utils import logging

import hashlib
import os
from typing import Optional, Any, TYPE_CHECKING

LOGGER = logging.get_logger(__name__)

if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec


def text_to_module(source: str, file_name: str = "anon", lang="qs") -> Any:
    """Text to module function."""

    from pineboolib.application.parsers.parser_qsa import flscriptparse, postparse, pytnyzer
    from pineboolib.application import file, PROJECT
    from importlib import util
    import sys as python_sys

    db_name = PROJECT.conn_manager.mainConn().DBName()
    source_bytes = source.encode()
    sha_ = hashlib.new("sha1", source_bytes).hexdigest()
    module_name = "%s" % (file_name)
    fileobj = file.File("anon", "%s.py" % module_name, sha_, db_name=db_name)
    file_name = fileobj.path()

    if not os.path.isfile(file_name) or not os.path.getsize(
        file_name
    ):  # Si no existe el fichero o está vacio.
        file_dir = os.path.dirname(file_name)
        if not os.path.exists(file_dir):  # Si no existe la carpeta la crea
            os.makedirs(file_dir)
        elif os.path.exists(file_name):  # Si existe la carpeta borra el archivo erroneo
            os.remove(file_name)
        if lang == "qs":
            prog = flscriptparse.parse(source)
            if prog is None:
                raise ValueError("Failed to convert to Python")
            tree_data = flscriptparse.calctree(prog, alias_mode=0)
            ast = postparse.post_parse(tree_data)

            file_ = open(file_name, "w", encoding="UTF-8")

            pytnyzer.write_python_file(file_, ast)
            file_.close()
        else:
            file_ = open(file_name, "w", encoding="UTF-8")
            file_.write(source)
            file_.close()

        LOGGER.debug("Nuevo módulo anónimo generado -> %s " % file_name)
    else:
        LOGGER.debug("Usando módulo anónimo ya existente -> %s" % file_name)

    module_path = "tempdata.%s" % (module_name)

    spec: Optional["ModuleSpec"] = util.spec_from_file_location(module_path, file_name)
    if spec and spec.loader is not None:
        module = util.module_from_spec(spec)
        python_sys.modules[spec.name] = module
        spec.loader.exec_module(module)  # type: ignore [attr-defined]
        return module
    else:
        raise Exception("Module named %s can't be loaded from %s" % (module_path, file_name))



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Flupdates_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Flupdates(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flupdates class."""

    __tablename__ = "flupdates"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flupdates",
        "alias": "Registro de instalación/actualización de paquetes",
        "fields": [
            {
                "name": "id",
                "alias": "ID",
                "pk": True,
                "type": "serial",
                "null": False,
                "visiblegrid": False,
                "editable": False,
            },
            {
                "name": "actual",
                "alias": "Instalado actualmente",
                "type": "bool",
                "null": False,
                "default": True,
                "editable": False,
            },
            {"name": "fecha", "alias": "Fecha", "type": "date", "null": False, "editable": False},
            {"name": "hora", "alias": "Hora", "type": "time", "null": False, "editable": False},
            {
                "name": "nombre",
                "alias": "Nombre del Paquete",
                "type": "string",
                "length": 255,
                "null": False,
                "editable": False,
            },
            {
                "name": "modulesdef",
                "alias": "Modules Def.",
                "type": "stringlist",
                "null": False,
                "visible": False,
                "visiblegrid": False,
                "editable": False,
            },
            {
                "name": "filesdef",
                "alias": "Files Def.",
                "type": "stringlist",
                "null": False,
                "visible": False,
                "visiblegrid": False,
                "editable": False,
            },
            {
                "name": "shaglobal",
                "alias": "SHA1 Global",
                "type": "string",
                "length": 255,
                "null": False,
                "visible": False,
                "visiblegrid": False,
                "editable": False,
            },
            {
                "name": "auxtxt",
                "alias": "Auxiliar Texto",
                "type": "stringlist",
                "visible": False,
                "visiblegrid": False,
                "editable": False,
            },
            {
                "name": "auxbin",
                "alias": "Auxiliar Binario",
                "type": "bytearray",
                "visible": False,
                "visiblegrid": False,
                "editable": False,
            },
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    id = sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True)
    actual = sqlalchemy.Column("actual", sqlalchemy.Boolean)
    fecha = sqlalchemy.Column("fecha", sqlalchemy.Date)
    hora = sqlalchemy.Column("hora", sqlalchemy.Time)
    nombre = sqlalchemy.Column("nombre", sqlalchemy.String(255))
    modulesdef = sqlalchemy.Column("modulesdef", sqlalchemy.String)
    filesdef = sqlalchemy.Column("filesdef", sqlalchemy.String)
    shaglobal = sqlalchemy.Column("shaglobal", sqlalchemy.String(255))
    auxtxt = sqlalchemy.Column("auxtxt", sqlalchemy.String)
    auxbin = sqlalchemy.Column("auxbin", sqlalchemy.LargeBinary)


# <--- Fields ---



# -*- coding: utf-8 -*-
"""
Main Eneboo-alike UI.
"""
from PyQt6 import QtWidgets, QtCore, QtGui, QtXml  # type: ignore[import]

from pineboolib.core import settings
from pineboolib.core.utils import utils_base


from pineboolib.fllegacy.aqsobjects import aqsobjectfactory

from pineboolib.fllegacy import flformdb, systype

from pineboolib import logging, application

from pineboolib.interfaces import imainwindow


from typing import Any, Dict, Optional, List, cast, Union


QSA_SYS = systype.SysType()
LOGGER = logging.get_logger(__name__)
AQS = aqsobjectfactory.AQS


class DockListView(QtCore.QObject):
    """DockListWiew class."""

    doc_widget: QtWidgets.QDockWidget
    tree_widget: QtWidgets.QTreeWidget
    action_group: QtGui.QActionGroup
    set_visible = QtCore.pyqtSignal(bool)
    Close = QtCore.pyqtSignal(bool)

    def __init__(self, parent=None, name: str = "dockListView", title: str = "") -> None:
        """Initialize the DockListView instance."""

        super(DockListView, self).__init__(parent)
        if parent is None:
            return

        self.setObjectName(name)

        self.doc_widget = QtWidgets.QDockWidget(name, parent)
        self.doc_widget.setObjectName("%sListView" % name)
        self.doc_widget.setFixedWidth(300)

        self.tree_widget = QtWidgets.QTreeWidget(self.doc_widget)
        self.tree_widget.setObjectName(self.objectName())
        self.tree_widget.setColumnCount(2)
        self.tree_widget.setHeaderLabels(["", ""])
        self.tree_widget.headerItem().setHidden(True)  # type: ignore [union-attr]
        self.tree_widget.hideColumn(1)

        self.doc_widget.setWidget(self.tree_widget)
        self.doc_widget.setWindowTitle(title)

        """
        w.resizeEnabled = true;
        w.closeMode = true;
        w.setFixedExtentWidth(300);
        """

        self.tree_widget.doubleClicked.connect(  # type: ignore [attr-defined] # noqa: F821
            self.activateAction
        )

    def writeState(self) -> None:
        """Save the state and geometry."""

        if self.doc_widget is None:
            raise Exception("self.doc_widget is empty!")

        settings = aqsobjectfactory.AQSettings()
        key = "MainWindow/%s/" % self.doc_widget.objectName()
        # FIXME
        # settings.writeEntry("%splace" % key, self.doc_widget.place())  # Donde está emplazado

        settings.writeEntry("%svisible" % key, self.doc_widget.isVisible())
        settings.writeEntry("%sx" % key, self.doc_widget.x())
        settings.writeEntry("%sy" % key, self.doc_widget.y())
        settings.writeEntry("%swidth" % key, self.doc_widget.width())
        settings.writeEntry("%sheight" % key, self.doc_widget.height())
        # FIXME
        # settings.writeEntry("%soffset", key, self.offset())
        # area = self.area()
        # settings.writeEntry("%sindex" % key, area.findDockWindow(self.doc_widget) if area else None)

    def readState(self) -> None:
        """Read the state and geometry."""

        if self.doc_widget is None:
            raise Exception("self.doc_widget is empty!")

        if self.tree_widget is None:
            raise Exception("self.tree_widget is empty!")

        settings = aqsobjectfactory.AQSettings()
        key = "MainWindow/%s/" % self.doc_widget.objectName()
        # FIXME
        # place = settings.readNumEntry("%splace" % key, AQS.InDock)
        # if place == AQS.OutSideDock:
        #    self.doc_widget.setFloating(True)
        #    self.doc_widget.move(settings.readNumEntry("%sx" % key, self.doc_widget.x()),
        #                 settings.readNumEntry("%sy" % key, self.doc_widget.y()))

        # self.doc_widget.offset = settings.readNumEntry("%soffset" % key, self.offset)
        # index = settings.readNumEntry("%sindex" % key, None)
        # FIXME
        # if index is not None:
        #    area = w.area()
        #    if area:
        #        area.moveDockWindow(w, index)

        width = settings.readNumEntry("%swidth" % key, self.doc_widget.width())
        height = settings.readNumEntry("%sheight" % key, self.doc_widget.height())
        self.tree_widget.resize(width, height)
        # self.doc_widget.resize(width, height)

        if not application.PROJECT.DGI.mobilePlatform():
            visible = settings.readBoolEntry("%svisible" % key, True)
            if visible:
                self.doc_widget.show()

            else:
                self.doc_widget.hide()

            self.set_visible.emit(not self.doc_widget.isHidden())
        else:
            self.doc_widget.hide()
            self.set_visible.emit(False)
            self.doc_widget.close()

    def initFromWidget(self, doc_widget) -> None:
        """Initialize the internal widget."""

        self.doc_widget = doc_widget
        self.tree_widget = doc_widget.widget()
        if self.tree_widget and hasattr(self.tree_widget, "doubleClicked"):
            self.tree_widget.doubleClicked.connect(  # type: ignore [attr-defined] # noqa: F821
                self.activateAction
            )

    def change_state(self, state: bool) -> None:
        """Change the display status."""
        if self.doc_widget is None:
            raise Exception("not initialized")
        if state:
            self.doc_widget.show()
        else:
            self.doc_widget.close()

    def activateAction(self, item) -> None:
        """Activate the action associated with the active item."""

        if item is None or not self.action_group:
            return

        action_name = item.sibling(item.row(), 1).data()
        if action_name == "":
            return

        action: QtGui.QAction = cast(
            QtGui.QAction, self.action_group.findChild(QtGui.QAction, action_name)
        )
        if action:
            action.triggered.emit()  # type: ignore [attr-defined] # noqa: F821

    def update(
        self, action_group: Optional[QtGui.QActionGroup] = None, reverse: bool = False
    ) -> None:
        """Update available items."""
        if not action_group:
            return

        self.action_group = action_group

        if not self.tree_widget:
            return

        self.tree_widget.clear()

        self.buildListView(
            self.tree_widget, AQS.toXml(self.action_group), self.action_group, reverse
        )

    def buildListView(self, parent_item, parent_element, action_group_, reverse: bool) -> None:
        """Build the tree of available options."""

        this_item = None
        node = (
            parent_element.lastChild().toElement()
            if reverse
            else parent_element.firstChild().toElement()
        )
        while not node.isNull():
            if node.attribute("objectName") in ("", "separator"):  # Pasamos de este
                node = (
                    node.previousSibling().toElement()
                    if reverse
                    else node.nextSibling().toElement()
                )
                continue
            class_name = node.attribute("class")
            if class_name.startswith("QAction"):
                if class_name == "QActionGroup":
                    act_group = action_group_.findChild(
                        QtGui.QActionGroup, node.attribute("objectName")
                    )
                    if act_group and not act_group.isVisible():
                        node = (
                            node.previousSibling().toElement()
                            if reverse
                            else node.nextSibling().toElement()
                        )
                        continue

                    group_name = node.attribute("objectName")
                    if (
                        group_name not in ("pinebooActionGroup")
                        and not group_name.endswith("Actions")
                        and not group_name.startswith(("pinebooAg"))
                    ) or group_name.endswith("MoreActions"):
                        this_item = QtWidgets.QTreeWidgetItem(parent_item)
                        this_item.setText(0, group_name)

                    else:
                        this_item = parent_item

                    self.buildListView(this_item, node, action_group_, reverse)
                    node = (
                        node.previousSibling().toElement()
                        if reverse
                        else node.nextSibling().toElement()
                    )
                    continue

                if node.attribute("objectName") not in (
                    "pinebooActionGroup",
                    "pinebooActionGroup_actiongroup_name",
                ):
                    action_name = node.attribute("objectName")
                    action = action_group_.findChild(QtGui.QAction, action_name)

                    action_group = action_group_.findChild(
                        QtGui.QActionGroup, parent_element.attribute("objectName")
                    )
                    if action_group is not None:  # ¢heck if is visible
                        ac_orig = action_group.findChild(QtGui.QAction, action_name)
                        if ac_orig and not ac_orig.isVisible():
                            action = None

                    if action is not None:
                        if action_name.endswith("actiongroup_name"):
                            this_item = (
                                parent_item
                                if isinstance(parent_item, QtWidgets.QTreeWidgetItem)
                                else None
                            )
                        else:
                            this_item = QtWidgets.QTreeWidgetItem(parent_item)

                        if this_item is not None:
                            this_item.setIcon(0, action.icon())
                            this_item.setText(1, action_name)
                            this_item.setText(0, node.attribute("text").replace("&", ""))

                self.buildListView(this_item, node, action_group_, reverse)

            node = node.previousSibling().toElement() if reverse else node.nextSibling().toElement()


class MainForm(imainwindow.IMainWindow):
    """
    Create Eneboo-alike UI.
    """

    MAX_RECENT = 10
    app_ = None
    ag_menu_: Optional["QtGui.QActionGroup"]
    ag_rec_: Optional["QtGui.QActionGroup"]
    ag_mar_: Optional["QtGui.QActionGroup"]
    dck_mod_: "DockListView"
    dck_rec_: "DockListView"
    dck_mar_: "DockListView"
    tab_widget: "QtWidgets.QTabWidget"
    w_: "QtWidgets.QMainWindow"
    # tw_corner = None  # deprecated
    act_sig_map_: "QtCore.QSignalMapper"
    initialized_mods_: List[str]

    main_widgets_: Dict[str, "QtWidgets.QWidget"] = {}
    # lista_tabs_ = []

    def __init__(self) -> None:
        """Construct Eneboo-alike UI."""
        super().__init__()
        self.main_widget = self
        self.ag_menu_ = None
        self.ag_rec_ = None
        self.ag_mar_ = None

    def eventFilter(
        self, obj_: Optional["QtCore.QObject"], event: Optional["QtCore.QEvent"]
    ) -> bool:
        """Process GUI events."""

        if isinstance(event, AQS.ContextMenu):
            if obj_ == getattr(self.dck_mod_, "doc_widget", None):
                return self.addMarkFromItem(
                    self.dck_mod_.tree_widget.currentItem(), event.globalPos()
                )
            elif obj_ == getattr(self.dck_rec_, "doc_widget", None):
                return self.addMarkFromItem(
                    self.dck_rec_.tree_widget.currentItem(), event.globalPos()
                )
            elif obj_ == getattr(self.dck_mar_, "doc_widget", None):
                return self.removeMarkFromItem(
                    self.dck_mar_.tree_widget.currentItem(), event.globalPos()
                )

            # pinebooMenu = self.main_widget.child("pinebooMenu")
            # pinebooMenu.exec_(e.globalPos)
            return True

        elif isinstance(event, AQS.Close):
            if isinstance(obj_, MainForm):
                if self.main_widget:
                    self.main_widget.setDisabled(True)
                ret = self.exit()
                if not ret:
                    if self.main_widget:
                        self.main_widget.setDisabled(False)
                    event.ignore()

                return True

            elif isinstance(obj_, QtWidgets.QDockWidget):
                cast(
                    QtCore.pyqtSignal, obj_.topLevelChanged
                ).emit(  # type: ignore [attr-defined] # noqa: F821
                    False
                )
            elif isinstance(obj_, flformdb.FLFormDB):
                for number in range(self.tab_widget.count()):
                    if self.tab_widget.widget(number) is obj_:
                        self.tab_widget.removeTab(number)
                        break

        elif isinstance(event, AQS.Show):
            if isinstance(obj_, flformdb.FLFormDB):
                return True

        return False

    def createUi(self, ui_file: str) -> None:
        """Create UI from file path."""

        mng = application.PROJECT.conn_manager.managerModules()
        self.main_widget = cast(QtWidgets.QMainWindow, mng.createUI(ui_file, None, self))
        self.main_widget.setObjectName("container")

    def exit(self) -> bool:
        """Process exit events."""
        do_exit = application.PROJECT.aq_app.queryExit()
        if do_exit:
            self.writeState()
            self.main_widget.removeEventFilter(self.main_widget)
            self.removeAllPages()

        return do_exit

    def writeStateModule(self) -> None:
        """Write settings for modules."""

        pass

    def writeState(self) -> None:
        """Save settings."""

        if self.dck_rec_ is None:
            raise Exception("Recent dockListView is missing!")

        if self.dck_mod_ is None:
            raise Exception("Modules dockListView is missing!")

        if self.dck_mar_ is None:
            raise Exception("BookMarks dockListView is missing!")

        self.dck_mod_.writeState()
        self.dck_rec_.writeState()
        self.dck_mar_.writeState()

        settings = aqsobjectfactory.AQSettings()
        key = "MainWindow/"

        settings.writeEntry("%smaximized" % key, self.isMaximized())
        settings.writeEntry("%sx" % key, self.x())
        settings.writeEntry("%sy" % key, self.y())
        settings.writeEntry("%swidth" % key, self.width())
        settings.writeEntry("%sheight" % key, self.height())

        key += "%s/" % application.PROJECT.conn_manager.database()

        open_actions = []

        for i in range(self.tab_widget.count()):
            open_actions.append(cast(flformdb.FLFormDB, self.tab_widget.widget(i)).idMDI())

        settings.writeEntryList("%sopenActions" % key, open_actions)
        settings.writeEntry("%scurrentPageIndex" % key, self.tab_widget.currentIndex())

        recent_actions = []
        root_recent = self.dck_rec_.tree_widget.invisibleRootItem()
        if root_recent:
            count_recent = root_recent.childCount()
            for i in range(count_recent):
                recent_actions.append(root_recent.child(i).text(1))  # type: ignore [union-attr]
        settings.writeEntryList("%srecentActions" % key, recent_actions)

        mark_actions = []
        root_mark = self.dck_mar_.tree_widget.invisibleRootItem()
        if root_mark:
            count_mark = root_mark.childCount()
            for i in range(count_mark):
                mark_actions.append(root_mark.child(i).text(1))  # type: ignore [union-attr]
        settings.writeEntryList("%smarkActions" % key, mark_actions)

    def readState(self) -> None:
        """Read settings."""

        if self.dck_rec_ is None:
            raise Exception("Recent dockListView is missing!")

        if self.dck_mod_ is None:
            raise Exception("Modules dockListView is missing!")

        if self.dck_mar_ is None:
            raise Exception("BookMarks dockListView is missing!")

        self.dck_mod_.readState()
        self.dck_rec_.readState()
        self.dck_mar_.readState()

        settings = aqsobjectfactory.AQSettings()
        key = "MainWindow/"

        maximized = settings.readBoolEntry("%smaximized" % key)

        if not maximized:
            pos_x = settings.readNumEntry("%sx" % key)
            pos_y = settings.readNumEntry("%sy" % key)
            if QSA_SYS.osName() == "MACX" and pos_y < 20:
                pos_y = 20
            self.move(pos_x, pos_y)
            self.resize(
                settings.readNumEntry("%swidth" % key, self.width()),
                settings.readNumEntry("%sheight" % key, self.height()),
            )
        else:
            self.main_widget.showMaximized()

        self.loadTabs()

    def loadTabs(self) -> None:
        """Load tabs."""
        if self.ag_menu_:
            settings = aqsobjectfactory.AQSettings()
            key = "MainWindow/%s/" % application.PROJECT.conn_manager.database()

            open_actions = settings.readListEntry("%sopenActions" % key)

            if self.tab_widget is None:
                raise Exception("tab_widget is empty!")

            for number in range(self.tab_widget.count()):
                widget = self.tab_widget.widget(number)
                if widget is not None:
                    widget.close()
                    self.tab_widget.removeTab(number)

            actions_opened: List[str] = []
            for open_action in open_actions:
                if open_action in actions_opened:
                    continue
                else:
                    actions_opened.append(open_action)

                action = cast(QtGui.QAction, self.ag_menu_.findChild(QtGui.QAction, open_action))
                if not action or not action.isVisible():
                    continue
                module_name = ""
                if action.objectName() in application.PROJECT.actions.keys():
                    ui_name = application.PROJECT.actions[action.objectName()]._master_form
                    if ui_name:
                        module_name = (
                            application.PROJECT.conn_manager.managerModules().idModuleOfFile(
                                "%s.ui" % ui_name
                            )
                        )
                if module_name:
                    self.initModule(module_name)

                self.addForm(open_action, action.icon().pixmap(16, 16))

            idx = settings.readNumEntry("%scurrentPageIndex" % key)
            if idx > 0 and idx < len(self.tab_widget):
                self.tab_widget.setCurrentWidget(self.tab_widget.widget(idx))

            recent_actions = settings.readListEntry("%srecentActions" % key)
            for recent in reversed(recent_actions):
                self.addRecent(cast(QtGui.QAction, self.ag_menu_.findChild(QtGui.QAction, recent)))

            mark_actions = settings.readListEntry("%smarkActions" % key)
            for mark in reversed(mark_actions):
                self.addMark(cast(QtGui.QAction, self.ag_menu_.findChild(QtGui.QAction, mark)))

    def init(self) -> None:
        """Initialize UI."""

        cast(QtWidgets.QMainWindow, self.main_widget).statusBar().hide()  # type: ignore [union-attr]
        self.main_widgets_ = {}
        self.act_sig_map_ = QtCore.QSignalMapper(self.main_widget)
        self.act_sig_map_.setObjectName("pinebooActSignalMap")
        self.act_sig_map_.mappedString.connect(self.triggerAction)  # type: ignore
        self.initTabWidget()
        self.initHelpMenu()
        self.initConfigMenu()
        self.initTextLabels()
        self.initDocks()
        self.initEventFilter()

    def initFromWidget(self, main_window: "QtWidgets.QMainWindow") -> None:
        """Initialize UI from a base widget."""
        self.main_widget = main_window
        self.main_widgets_ = {}
        self.act_sig_map_ = QtCore.QSignalMapper(self.main_widget)
        self.act_sig_map_.setObjectName("pinebooActSignalMap")
        self.tab_widget = cast(
            QtWidgets.QTabWidget, self.main_widget.findChild(QtWidgets.QTabWidget, "tabWidget")
        )
        self.action_group_menu = cast(
            QtGui.QActionGroup, self.main_widget.findChild(QtGui.QActionGroup, "pinebooActionGroup")
        )
        self.dck_mod_ = DockListView()
        self.dck_mod_.initFromWidget(
            cast(
                QtWidgets.QDockWidget,
                self.main_widget.findChild(QtWidgets.QDockWidget, "pinebooDockModulesListView"),
            )
        )
        self.dck_rec_ = DockListView()
        self.dck_rec_.initFromWidget(
            cast(
                QtWidgets.QDockWidget,
                self.main_widget.findChild(QtWidgets.QDockWidget, "pinebooDockRecentListView"),
            )
        )
        self.dck_mar_ = DockListView()
        self.dck_mar_.initFromWidget(
            cast(
                QtWidgets.QDockWidget,
                self.main_widget.findChild(QtWidgets.QDockWidget, "pinebooDockMarksListView"),
            )
        )
        self.initEventFilter()

    def initEventFilter(self) -> None:
        """Install event filters."""
        # w = self.main_widget
        # self.main_widget.eventFilterFunction = (
        #    "flapplication.flapplication.aqApp.Script.mainWindow_.eventFilter"
        # )

        # self.main_widget.allow_events = [AQS.ContextMenu, AQS.Close]

        self.main_widget.installEventFilter(self)
        if self.dck_mod_ and self.dck_mod_.doc_widget:
            self.dck_mod_.doc_widget.installEventFilter(self)
        if self.dck_rec_ and self.dck_rec_.doc_widget:
            self.dck_rec_.doc_widget.installEventFilter(self)
        if self.dck_mar_ and self.dck_mar_.doc_widget:
            self.dck_mar_.doc_widget.installEventFilter(self)

    def initModule(self, module: str) -> None:
        """Initialize main module."""
        if module in self.main_widgets_:
            mwi = self.main_widgets_[module]
            mwi.setObjectName(module)
            application.PROJECT.aq_app.setObjectName(module)
            mwi.show()

        if module not in self.initialized_mods_:
            self.initialized_mods_.append(module)
            application.PROJECT.call("%s.iface.init" % module, [], None, False)

        mng = application.PROJECT.conn_manager.managerModules()
        mng.setActiveIdModule(module)

    def removeCurrentPage(self, idx: Optional[int] = None) -> None:
        """Close tab."""
        if self.tab_widget is None:
            raise Exception("Not initialized.")

        if idx is None:
            for number in range(self.tab_widget.count()):
                if self.tab_widget.widget(number) is self.tab_widget.currentWidget():
                    idx = number
                    break

        if idx is not None:
            widget = self.tab_widget.widget(idx)
            if isinstance(widget, QtWidgets.QDialog):
                widget.close()
                # self.tab_widget.removeTab(idx)

    def removeAllPages(self) -> None:
        """Close all tabs."""
        if self.tab_widget is None:
            raise Exception("Not initialized.")

        # if len(tw):
        #    self.tab_widgetcorner.hide()

        for number in range(self.tab_widget.count()):
            widget = self.tab_widget.widget(number)
            if widget:
                widget.close()

    def addForm(self, action_str: str, icono: "QtGui.QPixmap") -> None:
        """Add new tab."""

        if self.tab_widget is None:
            raise Exception("tab_widget is empty!")

        action_name = action_str

        for i in range(self.tab_widget.count()):
            form = self.tab_widget.widget(i)
            if isinstance(form, flformdb.FLFormDB):
                if form.action().name() == action_name:
                    form.close()
                    break
        if self.ag_menu_:
            action = cast(QtGui.QAction, self.ag_menu_.findChild(QtGui.QAction, action_name))
            icon = action.icon() if action is not None else None
            try:
                form = aqsobjectfactory.AQFormDB(action_name, self.tab_widget)
                form.setMainWidget()

                title = form.windowTitle()

                if not form.mainWidget():
                    return
                if self.ag_menu_:
                    self.tab_widget.addTab(form, icon, title)

                form.setIdMDI(action_name)
                form.show()

                self.tab_widget.setCurrentWidget(form)
                form.installEventFilter(self.main_widget)
            except RuntimeError as error:
                LOGGER.warning(str(error))

    def addRecent(self, action: "QtGui.QAction") -> None:
        """Add new entry to recent list."""
        if not action:
            return

        if not self.ag_rec_:
            self.ag_rec_ = QtGui.QActionGroup(self.main_widget)

        check_max = True

        new_ag_rec_ = QtGui.QActionGroup(self.main_widget)
        new_ag_rec_.setObjectName("pinebooAgRec")

        self.cloneAction(action, new_ag_rec_)

        for item in self.ag_rec_.actions():
            if item.objectName() == action.objectName():
                check_max = False
                continue

            self.cloneAction(item, new_ag_rec_)

        self.ag_rec_ = new_ag_rec_
        if self.dck_rec_ is None:
            return
        tree_widget = self.dck_rec_.tree_widget
        if tree_widget is None:
            return
        if check_max and tree_widget.topLevelItemCount() >= self.MAX_RECENT:
            last_name = tree_widget.topLevelItem(tree_widget.topLevelItemCount() - 1).text(1)  # type: ignore [union-attr]
            action_ = cast(QtGui.QAction, self.ag_rec_.findChild(QtGui.QAction, last_name))
            if action_:
                self.ag_rec_.removeAction(action_)
                del action_

        self.dck_rec_.update(self.ag_rec_)

    def addMark(self, action: "QtGui.QAction") -> None:
        """Add new entry to Mark list."""
        if not action:
            return

        if not self.ag_mar_:
            self.ag_mar_ = QtGui.QActionGroup(self.main_widget)

        new_ag_mar = QtGui.QActionGroup(self.main_widget)
        new_ag_mar.setObjectName("pinebooAgMar")

        for item in self.ag_mar_.actions():
            if item.objectName() == action.objectName():
                continue

            self.cloneAction(item, new_ag_mar)

        self.cloneAction(action, new_ag_mar)

        self.ag_mar_ = new_ag_mar
        if self.dck_mar_:
            self.dck_mar_.update(self.ag_mar_, True)

    def addMarkFromItem(self, item: Any, pos: "QtCore.QPoint") -> bool:
        """Add a new item to the Bookmarks docket."""

        if not item:
            return False

        if item.text(1) is None:
            return True

        pop_menu = QtWidgets.QMenu()
        pop_menu.move(pos)
        pop_menu.addAction(self.tr("Añadir Marcadores"))
        res = pop_menu.exec()
        if res and self.ag_menu_ is not None:
            action = cast(QtGui.QAction, self.ag_menu_.findChild(QtGui.QAction, item.text(1)))
            if action and not action.objectName().endswith("actiongroup_name"):
                self.addMark(action)

        return True

    def removeMarkFromItem(self, item: Any, pos: "QtCore.QPoint") -> bool:
        """Add a new item to the Bookmarks docket."""
        if not item or not self.ag_mar_ or self.dck_mar_ is None:
            return False
        if (
            self.dck_mar_.tree_widget is None
            or self.dck_mar_.tree_widget.invisibleRootItem().childCount() == 0  # type: ignore [union-attr]
        ):
            return False
        if item.text(1) is None:
            return True

        pop_menu = QtWidgets.QMenu()
        pop_menu.move(pos)
        pop_menu.addAction(self.tr("Eliminar Marcador"))
        res = pop_menu.exec()
        if res:
            action = cast(QtGui.QAction, self.ag_mar_.findChild(QtGui.QAction, item.text(1)))
            if action and self.ag_mar_:
                self.ag_mar_.removeAction(action)
                del action
                self.dck_mar_.update(self.ag_mar_)

        return True

    def updateMenu(self, action_group: "QtGui.QActionGroup", menu: Any) -> None:
        """Update the modules menu with the available options."""

        for obj_ in action_group.children():
            o_name = obj_.objectName()
            if (
                not getattr(obj_, "isVisible", None)
                or not obj_.isVisible()  # type: ignore [attr-defined] # noqa: F821
            ):
                continue

            action = None
            if isinstance(obj_, QtGui.QActionGroup):
                action_obj = obj_.findChild(QtGui.QAction, "%s_actiongroup_name" % o_name)
                if action_obj is None:
                    action_obj = obj_.findChild(
                        QtGui.QAction, "%s_module_actiongroup_name" % o_name
                    )
                if action_obj is None:
                    action_obj = obj_.findChild(
                        QtGui.QAction,
                        "%sActions_actiongroup_name" % o_name.replace("ActionsMore", ""),
                    )

                if action_obj is not None:
                    if obj_.objectName().endswith("Actions") and not obj_.objectName().endswith(
                        "MoreActions"
                    ):
                        new_menu = menu
                    else:
                        new_menu = menu.addMenu(
                            action_obj.icon(),  # type: ignore [attr-defined] # noqa: F821
                            action_obj.text(),  # type: ignore [attr-defined] # noqa: F821
                        )
                        new_menu.triggered.connect(
                            action_obj.trigger  # type: ignore [attr-defined] # noqa: F821
                        )
                    self.updateMenu(obj_, new_menu)

            elif obj_.objectName().endswith("_actiongroup_name"):
                continue
            elif o_name == "separator":
                action = menu.addAction("")
                action.setSeparator(True)
            else:
                if isinstance(obj_, QtGui.QAction):
                    if self.ag_menu_:
                        obj_real = cast(
                            QtGui.QAction, self.ag_menu_.findChild(QtGui.QAction, obj_.objectName())
                        )
                        if obj_real is not None:
                            obj_ = obj_real  # Fix invalid QActions
                    action = menu.addAction(obj_.text())
                    action.setIcon(obj_.icon())
                    action.triggered.connect(obj_.trigger)
                    action.setVisible(obj_.isVisible())
                else:
                    continue

            if action is not None:
                action.setObjectName(o_name)

    def updateMenuAndDocks(self) -> None:
        """Update the main menu and dockers."""
        # FIXME: Duplicated piece of code
        self.updateActionGroup()
        pineboo_menu = cast(QtWidgets.QMenu, self.findChild(QtWidgets.QMenu, "menuPineboo"))
        pineboo_menu.clear()

        if self.ag_menu_ is None:
            raise Exception("ag_menu_ is empty!")

        self.updateMenu(self.ag_menu_, pineboo_menu)

        aq_app = application.PROJECT.aq_app

        aq_app.setMainWidget(self.main_widget)

        if self.ag_menu_ is None:
            raise Exception("ag_menu_ is empty!")

        if not self.ag_rec_:
            self.ag_rec_ = QtGui.QActionGroup(self.main_widget)

        if not self.ag_mar_:
            self.ag_mar_ = QtGui.QActionGroup(self.main_widget)

        if self.dck_rec_ is None:
            raise Exception("Recent dockListView is missing!")

        if self.dck_mod_ is None:
            raise Exception("Modules dockListView is missing!")

        if self.dck_mar_ is None:
            raise Exception("BookMarks dockListView is missing!")

        self.dck_mod_.update(self.ag_menu_)
        self.dck_rec_.update(self.ag_rec_)
        self.dck_mar_.update(self.ag_mar_)
        cast(
            QtGui.QAction, self.findChild(QtGui.QAction, "aboutQtAction")
        ).triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            aq_app.aboutQt
        )
        cast(
            QtGui.QAction, self.findChild(QtGui.QAction, "aboutPinebooAction")
        ).triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            aq_app.aboutPineboo
        )
        cast(
            QtGui.QAction, self.findChild(QtGui.QAction, "fontAction")
        ).triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            aq_app.chooseFont
        )
        cast(
            QtGui.QAction, self.findChild(QtWidgets.QMenu, "style")
        ).triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            aq_app.showStyles
        )
        cast(
            QtGui.QAction, self.findChild(QtGui.QAction, "helpIndexAction")
        ).triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            aq_app.helpIndex
        )
        cast(
            QtGui.QAction, self.findChild(QtGui.QAction, "urlPinebooAction")
        ).triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            aq_app.urlPineboo
        )

    def updateActionGroup(self) -> None:
        """Update the available actions."""

        if self.ag_menu_:
            list_ = self.ag_menu_.children()
            for obj in list_:
                if isinstance(obj, QtGui.QAction):
                    self.ag_menu_.removeAction(obj)
                    del obj

            self.ag_menu_.deleteLater()
            del self.ag_menu_

        self.ag_menu_ = QtGui.QActionGroup(self.main_widget)
        self.ag_menu_.setObjectName("pinebooActionGroup")
        ac_name = QtGui.QAction(self.ag_menu_)
        ac_name.setObjectName("pinebooActionGroup_actiongroup_name")
        ac_name.setText(self.tr("Menú"))

        mng = application.PROJECT.conn_manager.managerModules()
        areas = mng.listIdAreas()

        if self.act_sig_map_ is None:
            raise Exception("self.act_sig_map_ is empty!")

        for area in areas:
            if not QSA_SYS.isDebuggerEnabled() and area == "sys":
                break
            action_group = QtGui.QActionGroup(self.ag_menu_)
            action_group.setObjectName(area)
            ag_action = QtGui.QAction(action_group)
            ag_action.setObjectName("%s_actiongroup_name" % area)
            ag_action.setText(mng.idAreaToDescription(area))
            ag_action.setIcon(QtGui.QIcon(AQS.pixmap_fromMimeSource("folder.png")))
            modules = mng.listIdModules(area)
            for module in modules:
                if module == "sys" and QSA_SYS.isUserBuild():
                    continue
                action: Union[QtGui.QAction, QtGui.QActionGroup] = QtGui.QActionGroup(action_group)
                action.setObjectName(module)
                if QSA_SYS.isQuickBuild():
                    if module == "sys":
                        continue
                actions = self.widgetActions("%s.ui" % module, cast(QtWidgets.QWidget, action))

                if not actions:
                    # ac.setObjectName("")
                    action.deleteLater()
                    action = QtGui.QAction(action_group)
                    if action:
                        action.setObjectName(module)

                ac_action = QtGui.QAction(action)
                ac_action.setObjectName("%s_module_actiongroup_name" % module)
                ac_action.setText(mng.idModuleToDescription(module))
                ac_action.setIcon(self.iconSet16x16(mng.iconModule(module)))
                ac_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.act_sig_map_.map
                )
                self.act_sig_map_.setMapping(
                    ac_action, "triggered():initModule():%s_module_actiongroup_name" % module
                )
                if module == "sys" and area == "sys":
                    if QSA_SYS.isDebuggerMode():
                        static_load = QtGui.QAction(action_group)
                        static_load.setObjectName("staticLoaderSetupAction")
                        static_load.setText(self.tr("Configurar carga estática"))
                        static_load.setIcon(
                            QtGui.QIcon(AQS.pixmap_fromMimeSource("folder_update.png"))
                        )
                        static_load.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                            self.act_sig_map_.map
                        )
                        self.act_sig_map_.setMapping(
                            static_load,
                            "triggered():staticLoaderSetup():%s" % static_load.objectName(),
                        )

                        re_init = QtGui.QAction(action_group)
                        re_init.setObjectName("reinitAction")
                        re_init.setText(self.tr("Recargar scripts"))
                        re_init.setIcon(QtGui.QIcon(AQS.pixmap_fromMimeSource("reload.png")))
                        re_init.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                            self.act_sig_map_.map
                        )
                        self.act_sig_map_.setMapping(
                            re_init, "triggered():reinit():%s" % re_init.objectName()
                        )

        sh_console = QtGui.QAction(self.ag_menu_)
        sh_console.setObjectName("shConsoleAction")
        sh_console.setText(self.tr("Mostrar Consola de mensajes"))
        sh_console.setIcon(QtGui.QIcon(AQS.pixmap_fromMimeSource("consola.png")))
        sh_console.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            self.act_sig_map_.map
        )
        self.act_sig_map_.setMapping(
            sh_console, "triggered():shConsole():%s" % sh_console.objectName()
        )

        exit = QtGui.QAction(self.ag_menu_)
        exit.setObjectName("exitAction")
        exit.setText(self.tr("&Salir"))
        exit.setIcon(QtGui.QIcon(AQS.pixmap_fromMimeSource("exit.png")))
        exit.triggered.connect(self.act_sig_map_.map)  # type: ignore [attr-defined] # noqa: F821
        self.act_sig_map_.setMapping(exit, "triggered():exit():%s" % exit.objectName())

    def initTabWidget(self) -> None:
        """Initialize the TabWidget."""
        self.tab_widget = cast(
            QtWidgets.QTabWidget, self.main_widget.findChild(QtWidgets.QTabWidget, "tabWidget")
        )
        if self.tab_widget is None:
            raise Exception("no tabWidget found")
        self.tab_widget.setTabsClosable(True)
        self.tab_widget.tabCloseRequested[int].connect(self.removeCurrentPage)  # type: ignore
        self.tab_widget.removeTab(0)
        """
        tb = self.tw_corner = QToolButton(tw, "tabWidgetCorner")
        tb.autoRaise = False
        tb.setFixedSize(16, 16)
        tb.setIconSet(self.iconset16x16(AQS.pixmap_fromMimeSource("file_close.png")))
        tb.clicked.connect(self.removeCurrentPage)
        tw.setCornerWidget(tb, AQS.TopRight)
        AQS.toolTip_add(tb, self.tr("Cerrar pestaña"))
        tb.hide()
        """

    def initHelpMenu(self) -> None:
        """Initialize help menu."""

        about_qt = cast(QtGui.QAction, self.main_widget.findChild(QtGui.QAction, "aboutQtAction"))
        about_qt.setIcon(self.iconSet16x16(AQS.pixmap_fromMimeSource("aboutqt.png")))
        # about_qt.triggered.connect(flapplication.aqApp.aboutQt)

        about_pineboo = cast(
            QtGui.QAction, self.main_widget.findChild(QtGui.QAction, "aboutPinebooAction")
        )
        about_pineboo.setIcon(self.iconSet16x16(AQS.pixmap_fromMimeSource("pineboo-logo-32.png")))
        # about_pineboo.triggered.connect(flapplication.aqApp.aboutPineboo)

        help_index = cast(
            QtGui.QAction, self.main_widget.findChild(QtGui.QAction, "helpIndexAction")
        )
        help_index.setIcon(self.iconSet16x16(AQS.pixmap_fromMimeSource("help_index.png")))
        # help_index.triggered.connect(flapplication.aqApp.helpIndex)

        url_pineboo = cast(
            QtGui.QAction, self.main_widget.findChild(QtGui.QAction, "urlPinebooAction")
        )
        url_pineboo.setIcon(self.iconSet16x16(AQS.pixmap_fromMimeSource("pineboo-logo-32.png")))
        # url_pineboo.triggered.connect(flapplication.aqApp.urlPineboo)

    def initConfigMenu(self) -> None:
        """Initialize config menu."""
        font = cast(QtGui.QAction, self.main_widget.findChild(QtGui.QAction, "fontAction"))
        font.setIcon(self.iconSet16x16(AQS.pixmap_fromMimeSource("font.png")))
        # font.triggered.connect(flapplication.aqApp.chooseFont)

        style = cast(QtWidgets.QMenu, self.main_widget.findChild(QtWidgets.QMenu, "style"))

        application.PROJECT.aq_app.initStyles()
        style.setIcon(self.iconSet16x16(AQS.pixmap_fromMimeSource("estilo.png")))
        # style.triggered.connect(flapplication.aqApp.showStyles)

    def initTextLabels(self) -> None:
        """Initialize the tags in the mainForm base."""

        text_label = cast(QtWidgets.QLabel, self.main_widget.findChild(QtWidgets.QLabel, "tLabel"))
        text_label2 = cast(
            QtWidgets.QLabel, self.main_widget.findChild(QtWidgets.QLabel, "tLabel2")
        )
        texto = aqsobjectfactory.AQUtil.sqlSelect("flsettings", "valor", "flkey='verticalName'")
        if texto and texto != "False":
            text_label.setText(texto)

        if aqsobjectfactory.AQUtil.sqlSelect("flsettings", "valor", "flkey='PosInfo'") == "True":
            text_ = "%s@%s" % (QSA_SYS.nameUser(), QSA_SYS.nameBD())
            if QSA_SYS.osName() == "MACX":
                text_ += "     "

            text_label2.setText(text_)

    def initDocks(self) -> None:
        """Initialize the 3 available docks."""

        self.dck_mar_ = DockListView(self.main_widget, "pinebooDockMarks", self.tr("Marcadores"))
        self.main_widget.addDockWidget(AQS.DockLeft, self.dck_mar_.doc_widget)  # type: ignore [attr-defined]
        self.dck_rec_ = DockListView(self.main_widget, "pinebooDockRecent", self.tr("Recientes"))
        self.main_widget.addDockWidget(AQS.DockLeft, self.dck_rec_.doc_widget)  # type: ignore [attr-defined]
        self.dck_mod_ = DockListView(self.main_widget, "pinebooDockModules", self.tr("Módulos"))
        self.main_widget.addDockWidget(AQS.DockLeft, self.dck_mod_.doc_widget)  # type: ignore [attr-defined]

        window_menu = cast(
            QtWidgets.QMenu, self.main_widget.findChild(QtWidgets.QMenu, "windowMenu")
        )
        sub_menu = window_menu.addMenu(self.tr("&Vistas"))

        docks = cast(List[DockListView], self.main_widget.findChildren(DockListView))
        for dock in docks:
            action = sub_menu.addAction(dock.doc_widget.windowTitle())  # type: ignore [union-attr]
            if action:
                action.setCheckable(True)
                # FIXME: Comprobar si estoy visible o no
                # action.setChecked(dock.doc_widget.isVisible())
                dock.set_visible.connect(action.setChecked)
                action.triggered.connect(dock.change_state)  # type: ignore [attr-defined] # noqa: F821
                cast(
                    QtCore.pyqtSignal, dock.doc_widget.topLevelChanged
                ).connect(  # type: ignore [attr-defined] # noqa: F821
                    action.setChecked
                )
                # dock.doc_widget.Close.connect(action.setChecked)

    def cloneAction(self, old_action, parent) -> Any:
        """Clone one action into another."""

        new_action = QtGui.QAction(parent)
        new_action.setObjectName(old_action.objectName())
        new_action.setText(old_action.text())
        new_action.setStatusTip(old_action.statusTip())
        new_action.setToolTip(old_action.toolTip())
        new_action.setWhatsThis(old_action.whatsThis())
        new_action.setEnabled(old_action.isEnabled())
        new_action.setVisible(old_action.isVisible())
        new_action.triggered.connect(old_action.trigger)  # type: ignore [attr-defined] # noqa: F821
        new_action.toggled.connect(old_action.toggle)  # type: ignore [attr-defined] # noqa: F821
        if not old_action.icon().isNull():
            new_action.setIcon(self.iconSet16x16(old_action.icon().pixmap(16, 16)))

        return new_action

    def addWidgetActions(self, node, action_group, widget) -> None:
        """Add actions belonging to a widget."""

        # print("Loading actions", node, action_group, widget)

        actions = node.elementsByTagName("action")
        if len(actions) == 0:
            actions = node.elementsByTagName("addaction")
        hide_group = True

        actions_widgets: List[QtGui.QAction] = []
        if isinstance(widget, (QtWidgets.QToolBar, QtWidgets.QMenu)):
            actions_widgets = widget.actions()

        # print("Actions de", widget, actions_widgets)
        for i in range(len(actions)):
            item = actions.at(i).toElement()
            for action_widget in actions_widgets:
                # print("Comparando", action_widget.objectName(), "-", item.attribute("name"))
                if action_widget.objectName() != item.attribute("name"):
                    continue
                # action_widget = widget.findChild(QtGui.QAction, item.attribute("name"))
                # print("widget action", item.attribute("name"), action_widget, widget)
                # if action_widget is None:
                #    continue

                if action_widget.isVisible():
                    hide_group = False

                prev = item.previousSibling().toElement()
                if not prev.isNull() and prev.tagName() == "separator":
                    sep_ = action_group.addAction("separator")
                    sep_.setObjectName("separator")
                    sep_.setSeparator(True)
                    # actGroup.addSeparator()

                self.cloneAction(action_widget, action_group)
                break

        if hide_group:
            action_group.setVisible(False)

    def widgetActions(
        self, ui_file: str, parent: "QtWidgets.QWidget"
    ) -> Optional["QtGui.QActionGroup"]:
        """Collect the actions provided by a widget."""
        mng = application.PROJECT.conn_manager.managerModules()
        doc = QtXml.QDomDocument()
        content_cached = mng.contentCached(ui_file)
        if not content_cached or not doc.setContent(content_cached):
            if content_cached:
                LOGGER.warning("No se ha podido cargar %s" % (ui_file))
            return None

        widget = mng.createUI(ui_file)
        if widget is None:
            raise Exception("Failed to create UI from %r" % ui_file)
        if not isinstance(widget, QtWidgets.QMainWindow):
            if widget:
                self.main_widgets_[widget.objectName()] = widget

            return None

        widget.setObjectName(parent.objectName())

        if application.PROJECT.aq_app.acl_:
            application.PROJECT.aq_app.acl_.process(widget)

        # flapplication.aqApp.setMainWidget(widget)

        # if (QSA_SYS.isNebulaBuild()):
        #    widget.show()

        widget.hide()

        reduced = settings.CONFIG.value("ebcomportamiento/ActionsMenuRed", False)
        root = doc.documentElement().toElement()
        action_group = QtGui.QActionGroup(parent)
        action_group.setObjectName("%sActions" % parent.objectName())
        if root.attribute("version") == "3.3":
            bars = root.namedItem("toolbars").toElement()
            menu = root.namedItem("menubar").toElement()
            items = menu.elementsByTagName("item")

        else:
            widgets = root.elementsByTagName("widget")
            for item in range(widgets.size()):
                if widgets.item(item).toElement().attribute("class") == "QToolBar":
                    bars = widgets.item(item).toElement()
                elif widgets.item(item).toElement().attribute("class") == "QMenuBar":
                    menu_ = widgets.item(item)
                    items = menu_.toElement().elementsByTagName("widget")

        if not reduced:
            self.addWidgetActions(bars, action_group, widget.findChild(QtWidgets.QToolBar))

        if len(items) > 0:
            if not reduced:
                sep_ = action_group.addAction("separator")
                if sep_:
                    sep_.setObjectName("separator")
                    sep_.setSeparator(True)

                menu_ag = QtGui.QActionGroup(action_group)
                menu_ag.setObjectName("%sMore" % action_group.objectName())
                menu_ag_name = QtGui.QAction(menu_ag)
                menu_ag_name.setObjectName("%s_actiongroup_name" % action_group.objectName())
                menu_ag_name.setText(self.tr("Mas"))
                menu_ag_name.setIcon(QtGui.QIcon(AQS.pixmap_fromMimeSource("plus.png")))
            else:
                menu_ag = QtGui.QActionGroup(action_group)
                menu_ag.setObjectName(action_group.objectName())

            for i in range(len(items)):
                itn = items.at(i).toElement()
                if itn.parentNode().toElement().tagName() == "item":
                    continue

                sub_menu_ag = QtGui.QActionGroup(menu_ag)

                if root.attribute("version") == "3.3":
                    text = itn.attribute("text")
                else:
                    text = itn.text()

                name = itn.attribute("name")

                if not reduced:
                    sub_menu_ag.setObjectName("%sActions" % menu_ag.objectName())

                else:
                    sub_menu_ag.setObjectName(text)

                sub_menu_ag_name = QtGui.QAction(sub_menu_ag)
                sub_menu_ag_name.setObjectName("%s_actiongroup_name" % sub_menu_ag.objectName())
                sub_menu_ag_name.setText(QSA_SYS.toUnicode(text, "iso-8859-1"))

                self.addWidgetActions(itn, sub_menu_ag, widget.findChild(QtWidgets.QMenu, name))

        conns = root.namedItem("connections").toElement()
        connections = conns.elementsByTagName("connection")
        mapped_list: List[str] = []
        for i in range(connections.length()):
            itn = connections.at(i).toElement()
            sender = itn.namedItem("sender").toElement().text()
            action = action_group.findChild(QtGui.QAction, sender)
            if action:
                signal = itn.namedItem("signal").toElement().text()
                if signal in ["activated()", "triggered()"]:
                    signal_fix = "triggered"
                    signal = "triggered()"

                slot = itn.namedItem("slot").toElement().text()
                if self.act_sig_map_ is not None:
                    map_name = "%s:%s:%s" % (signal, slot, action.objectName())
                    if map_name not in mapped_list:
                        getattr(action, signal_fix).connect(self.act_sig_map_.map)
                        self.act_sig_map_.setMapping(action, map_name)
                        mapped_list.append(map_name)
                # getattr(ac, signal).connect(self.act_sig_map_.map)
                # ac.triggered.connect(self.triggerAction)

                # print("Guardando señales  %s:%s:%s de %s" % (signal, slot, ac.name, ac))

        # flapplication.aqApp.setMainWidget(None)
        widget.close()
        return action_group

    def iconSet16x16(self, pix: "QtGui.QPixmap") -> "QtGui.QIcon":
        """Reduce the size of a pixmap to 16 * 16."""

        img_ = QtGui.QImage(QtGui.QPixmap(pix))
        if not img_.isNull():
            img_ = img_.scaled(16, 16)
        ret = QtGui.QIcon(QtGui.QPixmap(img_))
        return ret

    def show(self) -> None:
        """Show the mainform."""

        super(MainForm, self).show()
        self.activateWindow()
        # self.setCaptionMainWidget()

    def initScript(self) -> None:
        """Startup process."""

        self.main_widget = self
        self.createUi(utils_base.filedir("plugins/mainform/eneboo/mainform.ui"))
        self.init()
        self.updateMenuAndDocks()
        self.initialized_mods_ = []
        self.initModule("sys")
        # self.show()
        self.readState()

    def reinitScript(self) -> None:
        """Re-start process."""

        self.writeState()
        self.removeAllPages()
        cast(
            QtGui.QAction, self.findChild(QtGui.QAction, "aboutQtAction")
        ).triggered.disconnect(  # type: ignore [attr-defined] # noqa: F821
            application.PROJECT.aq_app.aboutQt
        )
        cast(
            QtGui.QAction, self.findChild(QtGui.QAction, "aboutPinebooAction")
        ).triggered.disconnect(  # type: ignore [attr-defined] # noqa: F821
            application.PROJECT.aq_app.aboutPineboo
        )
        self.main_widget = self
        self.updateMenuAndDocks()
        self.initialized_mods_ = []
        self.initModule("sys")
        self.readState()

    def triggerAction(self, signature: str) -> None:
        """Start a process according to a given pattern."""

        sgt = signature.split(":")
        # ok = True
        if self.ag_menu_ is None:
            raise Exception("Not initialized")
        action: Optional[QtGui.QAction] = cast(
            QtGui.QAction, self.ag_menu_.findChild(QtGui.QAction, sgt[2])
        )

        if action is None:
            LOGGER.debug("triggerAction: Action not Found: %s" % signature)
            return

        action_name = action.objectName()

        signal = sgt[0]
        if signal == "triggered()":
            if not action.isVisible() or not action.isEnabled():
                return
        else:
            LOGGER.debug("triggerAction: Unhandled signal: %s" % signature)
            return

        fn_ = sgt[1]
        if fn_ == "initModule()":
            self.initModule(action_name.replace("_module_actiongroup_name", ""))

        elif fn_ == "openDefaultForm()":
            self.addRecent(action)
            if action_name in application.PROJECT.actions.keys():
                module_name = application.PROJECT.actions[  # type: ignore [union-attr] # noqa: F821
                    action_name
                ]._mod.module_name
                if module_name:
                    self.initModule(module_name)

            self.addForm(action_name, action.icon().pixmap(16, 16))

        elif fn_ == "execDefaultScript()":
            self.addRecent(action)
            if action_name in application.PROJECT.actions.keys():
                module_name = application.PROJECT.actions[  # type: ignore [union-attr] # noqa: F821
                    action_name
                ]._mod.module_name
                if module_name:
                    self.initModule(module_name)

                application.PROJECT.actions[action_name].execMainScript(action_name)

        elif fn_ == "loadModules()":
            QSA_SYS.loadModules()

        elif fn_ == "exportModules()":
            QSA_SYS.exportModules()

        elif fn_ == "importModules()":
            QSA_SYS.importModules()

        elif fn_ == "updatePineboo()":
            QSA_SYS.updatePineboo()

        elif fn_ == "staticLoaderSetup()":
            application.PROJECT.aq_app.staticLoaderSetup()

        elif fn_ == "reinit()":
            QSA_SYS.reinit()

        elif fn_ == "mrProper()":
            QSA_SYS.Mr_Proper()

        elif fn_ == "shConsole()":
            application.PROJECT.aq_app.showConsole()

        elif fn_ == "exit()":
            self.close()

        else:
            LOGGER.debug("tiggerAction: Unhandled slot : %s" % signature)

    def child(self, name: str) -> Optional[QtCore.QObject]:
        """Find a child widget."""

        return self.main_widget.findChild(QtWidgets.QWidget, name)

    def setCaptionMainWidget(self, value: str = "") -> None:
        """Set application title."""
        value = "- " + value
        self.setWindowTitle("Pineboo %s %s" % (application.PROJECT.load_version(), value))


# mainWindow: MainForm
# mainWindow = MainForm()



"""Numberedit module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets, QtGui  # type: ignore[import]
from pineboolib.q3widgets import qlineedit, qlabel, qhboxlayout

from typing import Any, SupportsFloat, SupportsInt, Union, cast


class NumberEdit(QtWidgets.QWidget):
    """
    NumberEdit class.
    """

    def __init__(self) -> None:
        """Inicialize."""

        super().__init__()

        self.line_edit = qlineedit.QLineEdit(self)
        self.label_line_edit = qlabel.QLabel(self)
        self.label_line_edit.setMinimumWidth(150)
        lay = qhboxlayout.QHBoxLayout(self)
        lay.addWidget(self.label_line_edit)
        lay.addWidget(self.line_edit)
        lay.setContentsMargins(0, 0, 0, 0)
        self.setLayout(lay)
        self.validator = QtGui.QDoubleValidator()
        self.line_edit.setValidator(self.validator)

    def getValue(self) -> str:
        """Return actual value."""

        return self.line_edit.getText()

    def setValue(self, value: Any) -> None:
        """Set value."""

        if value in ["", None]:
            return

        self.line_edit.setText(value)

    def getDecimals(self) -> int:
        """Return decimals."""

        return cast(QtGui.QDoubleValidator, self.line_edit.validator()).decimals()

    def setDecimals(self, decimals: Union[bytes, str, SupportsInt]) -> None:
        """Set decimals."""

        cast(QtGui.QDoubleValidator, self.line_edit.validator()).setDecimals(int(decimals))

    def setMinimum(self, min: Union[bytes, str, SupportsFloat]) -> None:
        """Set minimum value."""

        if min in ["", None]:
            return

        cast(QtGui.QDoubleValidator, self.line_edit.validator()).setBottom(float(min))

    def getMinimum(self) -> Union[int, float]:
        """Return minimum value."""

        return cast(QtGui.QDoubleValidator, self.line_edit.validator()).bottom()

    def getMaximum(self) -> Union[int, float]:
        """Return maximum value."""

        return cast(QtGui.QDoubleValidator, self.line_edit.validator()).top()

    def setMaximum(self, max: Union[bytes, str, SupportsFloat]) -> Any:
        """Set maximum value."""

        if max in ["", None]:
            return

        return cast(QtGui.QDoubleValidator, self.line_edit.validator()).setTop(float(max))

    def getLabel(self) -> str:
        """Return dialog label."""

        return self.label_line_edit.get_text()

    def setLabel(self, label: str) -> None:
        """Set dialog label."""

        self.label_line_edit.setText(label)

    label = property(getLabel, setLabel)
    value = property(getValue, setValue)
    decimals = property(getDecimals, setDecimals)
    mimimum = property(getMinimum, setMinimum)
    maximum = property(getMaximum, setMaximum)



"""Test_static_loader module."""

import unittest
import os
from pineboolib.core import settings
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.staticloader.tests import fixture_path


class TestStaticLoader(unittest.TestCase):
    """TestStaticLoader Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        from pineboolib.core.utils.utils_base import filedir

        db_name = "temp_db"
        dirs = [True, filedir("./application/staticloader/tests/fixtures")]
        settings.CONFIG.set_value(
            "StaticLoader/%s/enabled" % (db_name), True
        )  # Para activar carga estática
        settings.CONFIG.set_value("StaticLoader/%s/dirs" % db_name, dirs)
        init_testing()

    def test_script_overload(self) -> None:
        """Test script overload loader."""
        from pineboolib.qsa import qsa
        from pineboolib import application

        self.assertEqual(qsa.from_project("sys").saluda(), "Hola!")

        self.assertTrue(
            "sys" in application.PROJECT.actions.keys(),
            "Los actions disponibles son %s" % application.PROJECT.actions.keys(),
        )

        action = application.PROJECT.actions["sys"]
        script = application.load_script.load_script("sys.qs", action)
        self.assertEqual(script.saluda(), "Hola!")  # type: ignore [operator] # noqa: F821

    def test_static_loader(self) -> None:
        """Test script overloaded again."""
        from pineboolib.qsa import qsa

        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa.sys.loadModules(path, False))
        cursor = qsa.FLSqlCursor("flfiles")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("nombre", "clientes_model.py")
        cursor.setValueBuffer("idmodulo", "flfactppal")
        cursor.setValueBuffer("contenido", " ")
        cursor.setValueBuffer("sha", qsa.util.sha1(" "))
        self.assertTrue(cursor.commitBuffer())
        qsa.sys.reinit()

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()
        db_name = "temp_db"
        settings.CONFIG.set_value(
            "StaticLoader/%s/enabled" % (db_name), False
        )  # Para activar carga estática
        settings.CONFIG.set_value("StaticLoader/%s/dirs" % db_name, [])



"""Flscripteditor module."""

from pineboolib.q3widgets import qtextedit


class FLScriptEditor(qtextedit.QTextEdit):
    """FLScriptEditor class."""

    def __init__(self, name: str) -> None:
        """Inicialize."""

        super().__init__()

    def exec_(self):
        """Show edior."""

        self.show()

    def code(self) -> str:
        """Return text."""

        return self.PlainText



"""Custom widgets package."""

from typing import Tuple, Any

try:
    MODULE: Any
    pluginType = MODULE  # noqa : F821
except Exception:
    pass


def moduleInformation() -> Tuple[str, str]:  # pylint: disable=invalid-name
    """Return module inormation."""

    return "", ""



"""Test_pnsqlquery module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from PyQt6 import QtWidgets, QtCore, QtGui  # type: ignore[import]


class TestQT3UIParser(unittest.TestCase):
    """TestQT3UIParser Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_mainForm(self) -> None:
        """Test mainForm widget."""

        mng_modules = application.PROJECT.conn_manager.managerModules()
        from pineboolib.core.utils.utils_base import filedir

        file_1 = filedir("./application/parsers/parser_ui/tests/fixtures/main_form_qt3.ui")
        widget = mng_modules.createUI(file_1)
        self.assertTrue(widget)
        if widget:
            action = widget.findChild(QtGui.QAction, "ebcomportamiento")
        self.assertTrue(action)

    def test_formRecord(self) -> None:
        """Test formRecord widget."""

        mng_modules = application.PROJECT.conn_manager.managerModules()
        from pineboolib.core.utils.utils_base import filedir

        file_1 = filedir("./application/parsers/parser_ui/tests/fixtures/form_record_qt3.ui")
        widget = mng_modules.createUI(file_1)
        self.assertTrue(widget)
        if widget:
            bt_01 = widget.findChild(
                QtWidgets.QWidget, "pb_uno", QtCore.Qt.FindChildOption.FindChildrenRecursively
            )
        self.assertTrue(bt_01)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""QIconSet module."""

from PyQt6 import QtGui  # type: ignore[import]


class QIconSet(QtGui.QIcon):
    """QIconSet class."""

    def __init__(self, icon: QtGui.QIcon):
        """Initialize."""

        super().__init__(icon)



"""Groupbox module."""

# -*- coding: utf-8 -*-

from PyQt6 import QtWidgets  # type: ignore[import]
from pineboolib.q3widgets import qgroupbox


class GroupBox(qgroupbox.QGroupBox):
    """GroupBox class."""

    def __init__(self, *args) -> None:
        """Inicialize."""

        super().__init__(*args)
        self.setLayout(QtWidgets.QVBoxLayout())

    def add(self, widget: QtWidgets.QWidget) -> None:
        """Add new widget."""

        self.layout().addWidget(widget)  # type: ignore [union-attr]



#!/usr/bin/python3
"""
Simplify AST-XML structures for later generation of Python files.
"""
from importlib import machinery, util
from optparse import OptionParser
import os
import sys

from xml.etree import ElementTree as ET
from xml.dom import minidom  # type: ignore
from pineboolib import logging
from pineboolib.application.parsers.parser_qsa import pytnyzer, flscriptparse
from pineboolib.application.parsers import parser_qsa
from typing import List, Type, Optional, Dict, Tuple, Any, Callable, cast, Iterable


TreeData = Dict[str, Any]

LOGGER = logging.get_logger(__name__)

USEFUL_TOKENS = "ID,ICONST,FCONST,SCONST,CCONST,RXCONST,STATIC".split(",")

KNOWN_PARSERS: Dict[str, Type["TagObjectBase"]] = {}
UNKNOWN_PARSERS = {}


def parse_for(*tag_names: str) -> Callable:
    """Decorate functions for registering tags."""
    global KNOWN_PARSERS

    def decorator(func: Type["TagObjectBase"]) -> Type["TagObjectBase"]:
        for tag_name in tag_names:
            KNOWN_PARSERS[tag_name] = func
        return func

    return decorator


def parse(tag_name: str, tree_data: TreeData) -> "TagObject":
    """Excecute registered function for given tagname on treedata."""
    global KNOWN_PARSERS, UNKNOWN_PARSERS
    if tag_name not in KNOWN_PARSERS:
        UNKNOWN_PARSERS[tag_name] = 1
        func = parse_unknown
    else:
        func = KNOWN_PARSERS[tag_name]
    return func(tag_name, tree_data)


def getxmltagname(tag_name: str) -> str:
    """Transform tag names."""
    if tag_name == "source":
        return "Source"
    elif tag_name == "funcdeclaration":
        return "Function"
    elif tag_name == "classdeclaration":
        return "Class"
    elif tag_name == "vardeclaration":
        return "Variable"
    else:
        return "Unknown.%s" % tag_name


class TagObjectBase:
    """Base class for registering tag processors."""

    tags: List[str] = []

    @classmethod
    def can_process_tag(cls, tagname: str) -> bool:
        """Return if tagname is in class known tags."""
        return tagname in cls.tags

    def __init__(self, tagname: str) -> None:
        """Create base object for processing tags."""
        self.astname = tagname

    def add_subelem(self, argn: int, subelem: "TagObject") -> None:
        """Abstract function for adding sub elements."""

    def add_value(self, argn: int, vtype: str, value: str) -> None:
        """Abstract function for adding values."""

    def add_other(self, argn: int, vtype: str, data: str) -> None:
        """Abstract function for adding other types of data."""


XML_CLASS_TYPES: List[Type[TagObjectBase]] = []


class TagObjectFactory(type):
    """Metaclass for registering tag processors."""

    def __init__(cls, name: str, bases: Any, dct: Any) -> None:
        """Register a new class as tag processor."""
        global XML_CLASS_TYPES
        if issubclass(cls, TagObjectBase):
            XML_CLASS_TYPES.append(cast(Type[TagObjectBase], cls))
        else:
            raise Exception("This metaclass must be used as a subclass of TagObjectBase")
        super().__init__(name, bases, dct)


class TagObject(TagObjectBase, metaclass=TagObjectFactory):
    """Process XML tags for simplification. Main class with shared functionality."""

    set_child_argn = False
    name_is_first_id = False
    debug_other = True
    adopt_childs_tags: List[str] = []
    omit_tags = ["empty"]
    callback_subelem: Dict[Type["TagObject"], str] = {}
    promote_child_if_alone = False

    @classmethod
    def tagname(self, tagname: str) -> str:
        """Return processed target tag name."""
        return self.__name__

    def __init__(self, tagname: str) -> None:
        """Create new processor."""
        super().__init__(tagname)
        self.xml = ET.Element(self.tagname(tagname))
        self.xmlname: Optional[str] = None
        self.subelems: List[Any] = []
        self.values: List[Tuple[str, str]] = []
        if self.name_is_first_id:
            self.xml.set("name", "")

    def adopt_children(self, argn: int, subelem: "TagObject"):
        """Simplify tree by "merging" childs into itself."""
        for child in list(subelem.xml):
            if self.set_child_argn:
                child.set("argn", str(argn))
            else:
                if "argn" in child.attrib:
                    del child.attrib["argn"]
            self.xml.append(child)

    def omit_subelem(self, argn: int, subelem: "TagObject"):
        """Abstract function. Simplifies XML by removing unwanted terms."""
        return

    def is_in(self, listobj: Iterable) -> bool:
        """Return if the class type appears in any of the items."""
        return self.__class__ in listobj or self.astname in listobj

    def get(self, listobj: Dict[Any, str], default=None) -> Any:
        """Retrieve value from list based on this class type."""
        if self.__class__ in listobj:
            return listobj[self.__class__]
        if self.astname in listobj:
            return listobj[self.astname]
        return default

    def add_subelem(self, argn: int, subelem: "TagObject") -> None:
        """Add a new XML child."""
        if subelem.is_in(self.omit_tags):
            return self.omit_subelem(argn, subelem)
        if subelem.is_in(self.adopt_childs_tags):
            return self.adopt_children(argn, subelem)
        callback = subelem.get(self.callback_subelem)
        if callback:
            return getattr(self, callback)(argn, subelem)

        if self.set_child_argn:
            subelem.xml.set("argn", str(argn))
        self.xml.append(subelem.xml)
        self.subelems.append(subelem)

    def add_value(self, argn: int, vtype: str, value: str) -> None:
        """Add a new XML value."""
        self.values.append((vtype, value))
        if vtype == "ID" and self.name_is_first_id and self.xmlname is None:
            self.xmlname = value
            self.xml.set("name", value)
            return

        self.xml.set("arg%02d" % argn, vtype + ":" + repr(value))

    def add_other(self, argn: int, vtype: str, data: str) -> None:
        """Add extra data to XML."""
        if self.debug_other:
            self.xml.set("arg%02d" % argn, vtype)

    def polish(self) -> "TagObject":
        """Clean up the structure by removing or merging some data."""
        if self.promote_child_if_alone:
            if len(self.values) == 0 and len(self.subelems) == 1:
                return self.subelems[0]
        return self


class ListObject(TagObject):
    """Base class for list objects."""

    set_child_argn = False
    debug_other = False


class NamedObject(TagObject):
    """Base class for objects with names."""

    name_is_first_id = True
    debug_other = False


class ListNamedObject(TagObject):
    """Base class for list objects with names."""

    name_is_first_id = True
    set_child_argn = False
    debug_other = False


class TypedObject(ListObject):
    """Base class for typed objects."""

    type_arg = 0

    def add_other(self, argn, vtype, value):
        """Add extra data to XML."""
        if argn == self.type_arg:
            self.xml.set("type", vtype)


class Source(ListObject):
    """Process Source tags."""

    tags = ["source", "basicsource", "classdeclarationsource", "statement_list", "statement_block"]
    adopt_childs_tags = ["source_element", "statement_list", "statement", "statement_block"]


class Identifier(NamedObject):
    """Process Identifier tags."""

    tags = ["identifier", "optid"]

    def polish(self):
        """Fix astname attribute."""
        if self.xmlname is None:
            self.astname = "empty"
        return self


class Arguments(ListObject):
    """Process Argument tags."""

    tags = ["arglist"]
    adopt_childs_tags = ["vardecl_list"]


class VariableType(NamedObject):
    """Process VariableType tags."""

    tags = ["optvartype"]

    def polish(self):
        """Fix astname attribute."""
        if self.xmlname is None:
            self.astname = "empty"
        return self


class ExtendsType(NamedObject):
    """Process ExtendsType tags."""

    tags = ["optextends"]

    def polish(self):
        """Fix astname attribute."""
        if self.xmlname is None:
            self.astname = "empty"
        return self


class Function(ListNamedObject):
    """Process Function tags."""

    tags = ["funcdeclaration"]
    callback_subelem = ListNamedObject.callback_subelem.copy()
    callback_subelem[VariableType] = "add_vartype"

    def add_vartype(self, argn, subelem):
        """Add returns notation."""
        self.xml.set("returns", str(subelem.xmlname))


class FunctionAnon(ListObject):
    """Process FunctionAnon tags."""

    tags = ["funcdeclaration_anon"]


class FunctionAnonExec(ListObject):
    """Process FunctionAnonExec tags."""

    tags = ["funcdeclaration_anon_exec"]


class Variable(NamedObject):
    """Process Variable tags."""

    tags = ["vardecl"]
    callback_subelem = NamedObject.callback_subelem.copy()
    callback_subelem[VariableType] = "add_vartype"

    def add_vartype(self, argn, subelem):
        """Add type notation."""
        self.xml.set("type", str(subelem.xmlname))


class DeclarationBlock(ListObject):
    """Process DeclarationBlock tags."""

    tags = ["vardeclaration"]
    adopt_childs_tags = ["vardecl_list"]

    def add_other(self, argn, vtype, value):
        """Add debug info."""
        if argn == 0:
            self.xml.set("mode", vtype)

    def polish(self):
        """Cleanup."""
        # if len(self.values) == 0 and len(self.subelems) == 1:
        #    self.subelems[0].xml.set("mode",self.xml.get("mode"))
        #    return self.subelems[0]
        return self


class Class(ListNamedObject):
    """Process Class tags."""

    tags = ["classdeclaration"]
    callback_subelem = ListNamedObject.callback_subelem.copy()
    callback_subelem[ExtendsType] = "add_exttype"

    def add_exttype(self, argn, subelem):
        """Add extends notation."""
        self.xml.set("extends", str(subelem.xmlname))


class Member(TagObject):
    """Process Member tags."""

    debug_other = False
    set_child_argn = False
    tags = ["member_var", "member_call"]
    adopt_childs_tags = ["varmemcall", "member_var", "member_call"]


class ArrayMember(TagObject):
    """Process ArrayMember tags."""

    debug_other = False
    set_child_argn = False
    tags = ["array_member"]
    adopt_childs_tags = ["variable_1", "func_call"]


class InstructionCall(TagObject):
    """Process InstructionCall tags."""

    debug_other = False
    tags = ["callinstruction"]


class InstructionStore(TagObject):
    """Process InstructionStore tags."""

    promote_child_if_alone = True
    debug_other = False
    tags = ["storeinstruction"]


class InstructionFlow(TypedObject):
    """Process InstructionFlow tags."""

    debug_other = True
    tags = ["flowinstruction"]


class Instruction(TagObject):
    """Process Instruction tags."""

    promote_child_if_alone = True
    debug_other = False
    tags = ["instruction"]


class OpMath(TypedObject):
    """Process OpMath tags."""

    debug_other = True
    tags = ["mathoperator"]


class Compare(TypedObject):
    """Process Compare tags."""

    debug_other = True
    tags = ["cmp_symbol", "boolcmp_symbol"]


class FunctionCall(NamedObject):
    """Process FunctionCall tags."""

    tags = ["funccall_1"]


class CallArguments(ListObject):
    """Process CallArguments tags."""

    tags = ["callargs"]


class Constant(ListObject):
    """Process Constant tags."""

    tags = ["constant"]

    def add_value(self, argn: int, vtype: str, value: str) -> None:
        """Add value notation."""
        value = str(value)  # str(value,"ISO-8859-15","replace")
        if vtype == "SCONST":
            vtype = "String"
            value = value[1:-1]
            self.xml.set("delim", '"')
        if vtype == "CCONST":
            vtype = "String"
            value = value[1:-1]
            self.xml.set("delim", "'")
        if vtype == "RCONST":
            vtype = "Regex"
        if vtype == "ICONST":
            vtype = "Number"
        if vtype == "FCONST":
            vtype = "Number"
        self.const_value = value
        self.const_type = vtype
        self.xml.set("type", vtype)
        self.xml.set("value", value)


class InlineUpdate(ListObject):
    """Process InlineUpdate tags."""

    tags = ["inlinestoreinstruction"]

    def add_other(self, argn, vtype, value):
        """Add debug info."""
        self.xml.set("type", vtype)
        if argn == 0:
            self.xml.set("mode", "update-read")
        if argn == 1:
            self.xml.set("mode", "read-update")


class If(ListObject):
    """Process If tags."""

    tags = ["ifstatement"]


class Condition(ListObject):
    """Process Condition tags."""

    tags = ["condition"]


class TypeOf(ListObject):
    """Process Typeof_operator tags."""

    tags = ["typeof_operator"]


class Else(ListObject):
    """Process Else tags."""

    tags = ["optelse"]

    def polish(self):
        """Fix astname."""
        if len(self.subelems) == 0:
            self.astname = "empty"
        return self


class DictObject(ListObject):
    """Process DictObject tags."""

    tags = ["dictobject_value_elemlist", "dictobject_value"]
    adopt_childs_tags = ["dictobject_value_elemlist", "dictobject_value"]


class DictElem(ListObject):
    """Process DictElem tags."""

    tags = ["dictobject_value_elem"]


class ExpressionContainer(ListObject):
    """Process ExpressionContainer tags."""

    tags = ["expression"]
    # adopt_childs_tags = ['base_expression']

    def polish(self):
        """Fix internal expressions."""
        if len(self.values) == 0 and len(self.subelems) == 1:
            # if isinstance(self.subelems[0], Constant):
            if self.subelems[0].xml.tag == "base_expression":
                self.subelems[0].xml.tag = "Expression"
                return self.subelems[0]
            else:
                self.xml.tag = "Value"

        return self


class InstructionUpdate(ListObject):
    """Process InstructionUpdate tags."""

    tags = ["updateinstruction"]


class Switch(ListObject):
    """Process Switch tags."""

    tags = ["switch"]
    adopt_childs_tags = ["case_cblock_list", "case_block_list"]


class CaseList(ListObject):
    """Process CaseList tags."""

    tags = ["case_block_list"]
    adopt_childs_tags = ["case_cblock_list", "case_block_list"]


class Case(ListObject):
    """Process ExtendsType tags."""

    tags = ["case_block"]


class CaseDefault(ListObject):
    """Process CaseDefault tags."""

    tags = ["case_default"]


class While(ListObject):
    """Process While tags."""

    tags = ["whilestatement"]


class For(ListObject):
    """Process For tags."""

    tags = ["forstatement"]


class ForInitialize(ListObject):
    """Process ForInitialize tags."""

    tags = ["for_initialize"]


class ForCompare(ListObject):
    """Process ExtendsType tags."""

    tags = ["for_compare"]


class ForIncrement(ListObject):
    """Process ExtendsType tags."""

    tags = ["for_increment"]


class DoWhile(ListObject):
    """Process DoWhile tags."""

    tags = ["dowhilestatement"]


class ForIn(ListObject):
    """Process ExtendsType tags."""

    tags = ["forinstatement"]


class With(ListObject):
    """Process ExtendsType tags."""

    tags = ["withstatement"]


class TryCatch(ListObject):
    """Process TryCatch tags."""

    tags = ["trycatch"]


class New(ListObject):
    """Process New tags."""

    tags = ["new_operator"]


class Delete(ListObject):
    """Process Delete tags."""

    tags = ["deleteinstruction"]


class Parentheses(ListObject):
    """Process ExtendsType tags."""

    tags = ["parentheses"]
    adopt_childs_tags = ["base_expression"]


class OpUnary(TypedObject):
    """Process OpUnary tags."""

    tags = ["unary_operator"]


class OpTernary(ListObject):
    """Process OpTernary tags."""

    tags = ["ternary_operator"]


class OpUpdate(TypedObject):
    """Process OpUpdate tags."""

    tags = ["updateoperator"]


# ----- keep this one at the end.
class Unknown(TagObject):
    """Process Unknown tags."""

    promote_child_if_alone = True
    set_child_argn = False

    @classmethod
    def tagname(self, tagname):
        """Just return tagname."""
        return tagname

    @classmethod
    def can_process_tag(self, tagname):
        """Just return true."""
        return True


# -----------------


def create_xml(tagname) -> Optional[TagObject]:
    """Create processor for tagname by inspecting first known processor that fits."""
    classobj = None
    for cls in XML_CLASS_TYPES:
        if cls.can_process_tag(tagname):
            classobj = cls
            break
    if classobj is None:
        return None
    if issubclass(classobj, TagObject):
        return classobj(tagname)
    else:
        raise ValueError("Unexpected class %s" % classobj)


def parse_unknown(tagname, treedata):
    """Parse anything and error handling."""
    xmlelem = create_xml(tagname)
    if xmlelem is None:
        raise Exception("No class for parsing tagname %s" % tagname)

    position = 0
    for key, value in treedata["content"]:
        if type(value) is dict:
            instruction = parse(key, value)
            xmlelem.add_subelem(position, instruction)
        elif key in USEFUL_TOKENS:
            xmlelem.add_value(position, key, value)
        else:
            xmlelem.add_other(position, key, value)
        position += 1

    return xmlelem.polish()


def post_parse(treedata: TreeData):
    """Parse a xml. Convenience function."""
    source = parse("source", treedata)
    # print UNKNOWN_PARSERS.keys()
    return source.xml


class Module(object):
    """Python code tester for pineboo-parse."""

    def __init__(self, name: str, path: str) -> None:
        """Create Module."""
        self.name = name
        self.path = path

    def loadModule(self):
        """Import and return Python file."""
        try:
            module_name = self.name[: self.name.find(".")]
            script_name = os.path.join(self.path, self.name)
            spec: Optional["machinery.ModuleSpec"] = util.spec_from_file_location(
                module_name, script_name
            )
            if spec and spec.loader is not None:
                module = util.module_from_spec(spec)
                spec.loader.exec_module(module)  # type: ignore [attr-defined]
                self.module = module
            else:
                raise Exception(
                    "Module named %s can't be loaded from %s" % (module_name, script_name)
                )

            result = True

        except FileNotFoundError:
            LOGGER.error("Fichero %r no encontrado" % self.name)
            result = False
        except Exception:
            LOGGER.exception("Unexpected exception on loadModule")
            result = False

        return result


def parse_args(argv: List[str]) -> Tuple[Any, List[str]]:
    """Define parsing arguments for the program."""
    parser = OptionParser()
    parser.add_option(
        "-q",
        "--quiet",
        action="store_false",
        dest="verbose",
        default=True,
        help="don't print status messages to stdout",
    )

    parser.add_option(
        "--optdebug",
        action="store_true",
        dest="optdebug",
        default=False,
        help="debug optparse module",
    )

    parser.add_option(
        "--debug",
        action="store_true",
        dest="debug",
        default=False,
        help="prints lots of useless messages",
    )

    parser.add_option("--path", dest="storepath", default=None, help="store XML results in PATH")

    parser.add_option(
        "--topython",
        action="store_true",
        dest="topython",
        default=False,
        help="write python file from xml",
    )

    parser.add_option(
        "--exec-py",
        action="store_true",
        dest="exec_python",
        default=False,
        help="try to execute python file",
    )

    parser.add_option(
        "--toxml", action="store_true", dest="toxml", default=False, help="write xml file from qs"
    )

    parser.add_option(
        "--full", action="store_true", dest="full", default=False, help="write py file from qs"
    )

    parser.add_option(
        "--cache",
        action="store_true",
        dest="cache",
        default=False,
        help="If dest file exists, don't regenerate it",
    )

    parser.add_option(
        "--strict",
        action="store_true",
        dest="strict",
        default=False,
        help="Enable STRICT_MODE on pytnyzer",
    )

    parser.add_option(
        "--python-ext",
        dest="python_ext",
        default=".qs.py",
        help="Change Python file extension (default: '.qs.py')",
    )

    (options, args) = parser.parse_args(argv)
    return (options, args)


def main() -> None:
    """Run the program from command line."""

    log_format = "%(asctime)s - %(levelname)s: %(name)s: %(message)s"
    logging.basicConfig(format=log_format, level=0)
    blib_logger = logging.get_logger("blib2to3.pgen2.driver")
    blib_logger.setLevel(logging.WARNING)

    options, args = parse_args(sys.argv[1:])
    execute(options, args)


def pythonify(filelist: List[str], arguments: List[str] = []) -> None:
    """Convert to python the files included in the list."""
    if not isinstance(filelist, list):
        raise ValueError("First argument must be a list")
    options, args = parse_args(arguments)
    options.full = True
    execute(options, filelist)


def pythonify2(filename: str, known_refs: Dict[str, Tuple[str, str]] = {}) -> str:
    """Convert File to Python. Faster version as does not write to disk. Avoids re-parsing XML."""
    file_ = open(filename, "r", encoding="latin-1")
    filecontent = file_.read()
    file_.close()
    ast = common_parse(filecontent)
    return pytnyzer.pythonize2(ast, known_refs)


def pythonify_string(
    qs_code: str,
    known_refs: Dict[str, Tuple[str, str]] = {},
    parser_template: str = "expression_template",
) -> str:
    """Convert QS string to Python. For unit-testing, only evaluates expressions."""

    qs_code = flscriptparse.clean_no_python_tags(qs_code)

    ast = common_parse(qs_code)
    ast.set("parser-template", parser_template)  # type: ignore [attr-defined] # noqa: F821
    return pytnyzer.pythonize2(ast, known_refs)


def common_parse(data: str):
    """Run common parse code."""

    prog = flscriptparse.parse(data)
    if not prog:
        raise Exception("Parse failed")
    if prog["error_count"] > 0:
        raise Exception("Found %d errors parsing string" % (prog["error_count"]))

    tree_data: TreeData = flscriptparse.calctree(prog, alias_mode=0)
    return post_parse(tree_data)


def execute(options: Any, args: List[str]) -> None:
    """Execute conversion orders given by options and args. Can be used to emulate program calls."""

    parser_qsa.STRICT_MODE = options.strict

    if options.full:
        execpython = options.exec_python
        options.exec_python = False
        options.full = False
        options.toxml = True
        LOGGER.info("Pass 1 - Parse and write XML file . . .")
        try:
            execute(options, args)
        except Exception:
            LOGGER.exception("Error parseando:")

        options.toxml = False
        options.topython = True
        LOGGER.info("Pass 2 - Pythonize and write PY file . . .")
        try:
            execute(options, [arg + ".xml" for arg in args])
        except Exception:
            LOGGER.exception("Error convirtiendo:")

        if execpython:
            options.exec_python = execpython
            LOGGER.info("Pass 3 - Test PY file load . . .")
            options.topython = False
            try:
                execute(
                    options,
                    [(arg + ".xml.py").replace(".qs.xml.py", options.python_ext) for arg in args],
                )
            except Exception:
                LOGGER.exception("Error al ejecutar Python:")
        LOGGER.debug("Done.")

    elif options.exec_python:
        # import qsatype
        for filename in args:
            realpath = os.path.realpath(filename)
            path, name = os.path.split(realpath)
            if not os.path.exists(realpath):
                LOGGER.error("Fichero no existe: %s" % name)
                continue

            mod = Module(name, path)
            if not mod.loadModule():
                LOGGER.error("Error cargando modulo %s" % name)

    elif options.topython:
        import io

        if options.cache:
            args = [
                x
                for x in args
                if not os.path.exists((x + ".py").replace(".qs.xml.py", options.python_ext))
                or os.path.getmtime(x)
                > os.path.getctime((x + ".py").replace(".qs.xml.py", options.python_ext))
            ]

        nfs = len(args)
        for nf_, filename in enumerate(args):
            bname = os.path.basename(filename)
            if options.storepath:
                destname = os.path.join(options.storepath, bname + ".py")
            else:
                destname = filename + ".py"
            destname = destname.replace(".qs.xml.py", options.python_ext)
            if not os.path.exists(filename):
                LOGGER.error("Fichero %r no encontrado" % filename)
                continue
            LOGGER.debug(
                "Pythonizing File: %-35s . . . .        (%.1f%%)"
                % (bname, 100.0 * (nf_ + 1.0) / nfs)
            )
            old_stderr = sys.stdout
            stream = io.StringIO()
            sys.stdout = stream
            try:
                pytnyzer.pythonize(filename, destname, destname + ".debug")
            except Exception:
                LOGGER.exception("Error al pythonificar %r:" % filename)
            sys.stdout = old_stderr
            text = stream.getvalue()
            if len(text) > 2:
                LOGGER.info("%s: " % bname + ("\n%s: " % bname).join(text.splitlines()))

    else:
        if options.cache:
            args = [
                x
                for x in args
                if not os.path.exists(x + ".xml")
                or os.path.getmtime(x) > os.path.getctime(x + ".xml")
            ]
        nfs = len(args)
        for nf_, filename in enumerate(args):
            bname = os.path.basename(filename)
            LOGGER.debug(
                "Parsing File: %-35s . . . .        (%.1f%%)" % (bname, 100.0 * (nf_ + 1.0) / nfs)
            )
            try:
                file_ = open(filename, "r", encoding="latin-1")
                filecontent = file_.read()
                file_.close()
            except Exception:
                LOGGER.exception("Error: No se pudo abrir fichero %s", filename)
                continue
            prog = flscriptparse.parse(filecontent)
            if not prog:
                LOGGER.error("Error: No se pudo abrir %s" % (repr(filename)))
                continue
            if prog["error_count"] > 0:
                LOGGER.error(
                    "Encontramos %d errores parseando: %-35s"
                    % (prog["error_count"], repr(filename))
                )
                continue
            if not options.toxml:
                # Si no se quiere guardar resultado, no hace falta calcular mas
                continue

            tree_data = None
            try:
                tree_data = flscriptparse.calctree(prog, alias_mode=0)
            except Exception:
                LOGGER.exception("Error al convertir a XML %r:" % bname)

            if not tree_data:
                LOGGER.error("No se pudo parsear %s" % (repr(filename)))
                continue
            ast = post_parse(tree_data)
            if ast is None:
                LOGGER.error("No se pudo analizar %s" % (repr(filename)))
                continue
            if options.storepath:
                destname = os.path.join(options.storepath, bname + ".xml")
            else:
                destname = filename + ".xml"

            xml_str = minidom.parseString(ET.tostring(ast)).toprettyxml(indent="   ")
            file_ = open(destname, "w", encoding="UTF-8")
            file_.write(xml_str)
            file_.close()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""FLlarge_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Fllarge(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Fllarge class."""

    __tablename__ = "fllarge"

    # --- Metadata --->
    legacy_metadata = {
        "name": "fllarge",
        "alias": "Ficheros de gran tamaño",
        "fields": [
            {
                "name": "refkey",
                "alias": "Clave de Referencia",
                "pk": True,
                "type": "string",
                "length": 100,
                "null": False,
            },
            {"name": "sha1", "alias": "Hash Sha1", "type": "string", "length": 50},
            {"name": "contenido", "alias": "Contenido", "type": "stringlist", "visiblegrid": False},
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    refkey = sqlalchemy.Column("refkey", sqlalchemy.String(100), primary_key=True)
    sha1 = sqlalchemy.Column("sha1", sqlalchemy.String(50))
    contenido = sqlalchemy.Column("contenido", sqlalchemy.String)


# <--- Fields ---



"""
Core package for utilities.

This package holds all functions and classes that are like side utilities that don't require
any dependency from other folders. So they're safe to import.
"""

DISABLE_CHECK_MEMORY_LEAKS: bool = True  # Disabled memory leaks checking.



# -*- coding: utf-8 -*-
"""Manage a group by from a Query instance."""


class PNGroupByQuery(object):
    """Class that manages each group by of a query."""

    level_: int
    field_: str

    def __init__(self, label: int, field: str) -> None:
        """Specify the level and field."""

        self.level_ = label
        self.field_ = field

    def level(self) -> int:
        """Return the level GroupBy."""

        return self.level_

    def field(self) -> str:
        """Return the field GroupBy."""
        return self.field_



"""Ebcomportamiento module."""
# -*- coding: utf-8 -*-
from pineboolib import application
from pineboolib.qsa import qsa
from pineboolib.core import settings
from PyQt6 import QtWidgets, QtCore  # type: ignore[import]

import os

from typing import Any, Union, Optional


class FormInternalObj(qsa.FormDBWidget):
    """FormInternalObj class."""

    def main(self) -> None:
        """Entry function."""

        mng = qsa.aqApp.db().managerModules()
        self.ui_: "QtWidgets.QWidget" = mng.createUI(  # type: ignore [assignment]
            "ebcomportamiento.ui"
        )
        btn_accept = self.ui_.findChild(QtWidgets.QWidget, "pbnAceptar")  # type: ignore [attr-defined]
        btn_accept_tmp = self.ui_.findChild(QtWidgets.QWidget, "pbn_temporales")  # type: ignore [attr-defined]
        btn_cancel = self.ui_.findChild(QtWidgets.QWidget, "pbnCancelar")  # type: ignore [attr-defined]
        btn_color = self.ui_.findChild(QtWidgets.QWidget, "pbnCO")  # type: ignore [attr-defined]
        self.module_connect(btn_accept, "clicked()", self, "guardar_clicked")
        self.module_connect(btn_cancel, "clicked()", self, "cerrar_clicked")
        self.module_connect(btn_color, "clicked()", self, "color_chooser_clicked")
        self.module_connect(btn_accept_tmp, "clicked()", self, "cambiar_temporales_clicked")
        self.load_config()
        self.initEventFilter()
        if qsa.sys.interactiveGUI() == "Pineboo":
            self.ui_.show()

    def load_config(self) -> None:
        """Load configuration."""

        self.ui_.findChild(QtWidgets.QWidget, "cbFLTableDC").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("FLTableDoubleClick")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbFLTableSC").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("FLTableShortCut")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbFLTableCalc").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("FLTableExport2Calc")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbDebuggerMode").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("isDebuggerMode")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbSLConsola").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("SLConsola")
        )
        self.ui_.findChild(QtWidgets.QWidget, "leCallFunction").setText(  # type: ignore [attr-defined]
            self.read_local_value("ebCallFunction")
        )
        self.ui_.findChild(QtWidgets.QWidget, "leMaxPixImages").setText(  # type: ignore [attr-defined]
            self.read_local_value("maxPixImages")
        )
        self.ui_.findChild(QtWidgets.QWidget, "leNombreVertical").setText(  # type: ignore [attr-defined]
            self.read_db_value("verticalName")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbFLLarge").setChecked(  # type: ignore [attr-defined]
            self.read_db_value("FLLargeMode") == "True"
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbPosInfo").setChecked(  # type: ignore [attr-defined]
            self.read_db_value("PosInfo") == "True"
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbMobile").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("mobileMode")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbDeleteCache").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("deleteCache")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbParseProject").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("parseProject")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbNoPythonCache").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("noPythonCache")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbActionsMenuRed").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("ActionsMenuRed")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbSpacerLegacy").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("spacerLegacy")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cbParseModulesOnLoad").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("parseModulesOnLoad")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cb_traducciones").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("translations_from_qm")
        )
        self.ui_.findChild(QtWidgets.QWidget, "le_temporales").setText(  # type: ignore [attr-defined]
            self.read_local_value("temp_dir")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cb_kut_debug").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("kugar_debug_mode")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cb_no_borrar_cache").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("keep_general_cache")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cb_snapshot").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("show_snaptshop_button")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cb_imagenes").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("no_img_cached")
        )
        self.ui_.findChild(QtWidgets.QWidget, "cb_dbadmin").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("dbadmin_enabled")
        )

        self.ui_.findChild(QtWidgets.QWidget, "cb_preping").setChecked(  # type: ignore [attr-defined]
            self.read_local_value("preping")
        )

        valor = self.read_local_value("autoComp")
        auto_complete = "Siempre"
        if not valor or valor == "OnDemandF4":
            auto_complete = "Bajo Demanda (F4)"
        elif valor == "NeverAuto":
            auto_complete = "Nunca"

        self.ui_.findChild(  # type: ignore [attr-defined]
            QtWidgets.QWidget, "cbAutoComp"
        ).setCurrentText = auto_complete  # type: ignore [attr-defined]

        self.ui_.findChild(QtWidgets.QWidget, "leCO").hide()  # type: ignore [attr-defined]
        self.color_actual = self.read_local_value("colorObligatorio")
        if not self.color_actual:
            self.color_actual = "#FFE9AD"

        self.ui_.findChild(QtWidgets.QWidget, "leCO").setStyleSheet(  # type: ignore [attr-defined]
            "background-color:" + self.color_actual
        )

        self.ui_.findChild(QtWidgets.QWidget, "tbwLocales").setTabEnabled(5, False)  # type: ignore [attr-defined]

        self.ui_.findChild(QtWidgets.QWidget, "leCO").show()  # type: ignore [attr-defined]

    def read_db_value(self, valor_name: Optional[str] = None) -> Any:
        """Return global value."""
        util = qsa.FLUtil()
        value = util.sqlSelect("flsettings", "valor", "flkey='%s'" % valor_name)

        if value is None or valor_name == "verticalName" and isinstance(value, bool):
            value = ""

        return value

    def write_db_value(self, valor_name: str, value: Union[str, bool]) -> None:
        """Set global value."""
        util = qsa.FLUtil()
        if not util.sqlSelect("flsettings", "flkey", "flkey='%s'" % valor_name):
            util.sqlInsert("flsettings", "flkey,valor", "%s,%s" % (valor_name, value))
        else:
            util.sqlUpdate("flsettings", "valor", str(value), "flkey = '%s'" % valor_name)

    def read_local_value(self, valor_name: str) -> Any:
        """Return local value."""

        if valor_name in ("isDebuggerMode", "dbadmin_enabled"):
            valor = settings.CONFIG.value("application/%s" % valor_name, False)
        else:
            if valor_name in (
                "ebCallFunction",
                "maxPixImages",
                "kugarParser",
                "colorObligatorio",
                "temp_dir",
            ):
                valor = settings.CONFIG.value("ebcomportamiento/%s" % valor_name, "")
                if valor_name == "temp_dir" and valor == "":
                    app_ = qsa.aqApp
                    if app_ is None:
                        return ""

                    valor = app_.tmp_dir()

            else:
                valor = settings.CONFIG.value("ebcomportamiento/%s" % valor_name, False)
        return valor

    def write_local_value(self, valor_name: str, value: Union[str, bool]) -> None:
        """Set local value."""

        if valor_name in ("isDebuggerMode", "dbadmin_enabled"):
            settings.CONFIG.set_value("application/%s" % valor_name, value)
        else:
            if valor_name == "maxPixImages" and value is None:
                value = 600
            settings.CONFIG.set_value("ebcomportamiento/%s" % valor_name, value)

    def initEventFilter(self) -> None:
        """Inicialize event filter."""

        self.ui_.eventFilterFunction = qsa.ustr(  # type: ignore [attr-defined]
            self.ui_.objectName(), ".eventFilter"
        )
        self.ui_.allowedEvents = qsa.Array([qsa.AQS.Close])  # type: ignore [attr-defined]
        self.ui_.installEventFilter(self.ui_)

    def eventFilter(
        self, obj: Optional["QtCore.QObject"], event: Optional["QtCore.QEvent"]
    ) -> bool:
        """Event filter."""
        if type(event) == qsa.AQS.Close:  # noqa: E721
            self.cerrar_clicked()

        return True

    def cerrar_clicked(self) -> None:
        """Close the widget."""
        self.ui_.close()

    def guardar_clicked(self) -> None:
        """Save actual configuration."""

        self.write_db_value(
            "verticalName", self.ui_.findChild(QtWidgets.QWidget, "leNombreVertical").text()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "FLTableDoubleClick", self.ui_.findChild(QtWidgets.QWidget, "cbFLTableDC").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "FLTableShortCut", self.ui_.findChild(QtWidgets.QWidget, "cbFLTableSC").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "FLTableExport2Calc",
            self.ui_.findChild(QtWidgets.QWidget, "cbFLTableCalc").isChecked(),  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "isDebuggerMode", self.ui_.findChild(QtWidgets.QWidget, "cbDebuggerMode").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "SLConsola", self.ui_.findChild(QtWidgets.QWidget, "cbSLConsola").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "ebCallFunction", self.ui_.findChild(QtWidgets.QWidget, "leCallFunction").text()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "maxPixImages", self.ui_.findChild(QtWidgets.QWidget, "leMaxPixImages").text()  # type: ignore [attr-defined]
        )
        self.write_local_value("colorObligatorio", self.color_actual)
        self.write_local_value(
            "ActionsMenuRed", self.ui_.findChild(QtWidgets.QWidget, "cbActionsMenuRed").isChecked()  # type: ignore [attr-defined]
        )
        self.write_db_value(
            "FLLargeMode", self.ui_.findChild(QtWidgets.QWidget, "cbFLLarge").isChecked()  # type: ignore [attr-defined]
        )
        self.write_db_value(
            "PosInfo", self.ui_.findChild(QtWidgets.QWidget, "cbPosInfo").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "deleteCache", self.ui_.findChild(QtWidgets.QWidget, "cbDeleteCache").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "parseProject", self.ui_.findChild(QtWidgets.QWidget, "cbParseProject").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "noPythonCache", self.ui_.findChild(QtWidgets.QWidget, "cbNoPythonCache").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "mobileMode", self.ui_.findChild(QtWidgets.QWidget, "cbMobile").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "spacerLegacy", self.ui_.findChild(QtWidgets.QWidget, "cbSpacerLegacy").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "parseModulesOnLoad",
            self.ui_.findChild(QtWidgets.QWidget, "cbParseModulesOnLoad").isChecked(),  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "translations_from_qm",
            self.ui_.findChild(QtWidgets.QWidget, "cb_traducciones").isChecked(),  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "temp_dir", self.ui_.findChild(QtWidgets.QWidget, "le_temporales").text()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "kugar_debug_mode", self.ui_.findChild(QtWidgets.QWidget, "cb_kut_debug").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "keep_general_cache",
            self.ui_.findChild(QtWidgets.QWidget, "cb_no_borrar_cache").isChecked(),  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "show_snaptshop_button",
            self.ui_.findChild(QtWidgets.QWidget, "cb_snapshot").isChecked(),  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "no_img_cached", self.ui_.findChild(QtWidgets.QWidget, "cb_imagenes").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "dbadmin_enabled", self.ui_.findChild(QtWidgets.QWidget, "cb_dbadmin").isChecked()  # type: ignore [attr-defined]
        )
        self.write_local_value(
            "preping", self.ui_.findChild(QtWidgets.QWidget, "cb_preping").isChecked()  # type: ignore [attr-defined]
        )

        valor = self.ui_.findChild(QtWidgets.QWidget, "cbAutoComp").currentText()  # type: ignore [attr-defined]
        auto_complete = "AlwaysAuto"
        if valor == "Nunca":
            auto_complete = "NeverAuto"
        elif valor == "Bajo Demanda (F4)":
            auto_complete = "OnDemandF4"

        self.write_local_value("autoComp", auto_complete)
        self.cerrar_clicked()

    def color_chooser_clicked(self) -> None:
        """Set mandatory color."""
        self.color_actual = qsa.AQS.ColorDialog_getColor(self.color_actual, self.ui_).name()
        self.ui_.findChild(QtWidgets.QWidget, "leCO").setStyleSheet(  # type: ignore [attr-defined]
            "background-color:" + self.color_actual
        )

    def cambiar_temporales_clicked(self) -> None:
        """Change temp folder."""
        old_dir = self.ui_.findChild(QtWidgets.QWidget, "le_temporales").text()  # type: ignore [attr-defined]
        old_dir = os.path.normcase(old_dir)
        new_dir = qsa.FileDialog.getExistingDirectory(old_dir)
        if new_dir and new_dir is not old_dir:
            new_dir = new_dir[:-1]
            self.ui_.findChild(QtWidgets.QWidget, "le_temporales").setText(new_dir)  # type: ignore [attr-defined]

            application.PROJECT.tmpdir = new_dir



"""Flsqls module."""

from pineboolib import logging

from pineboolib.plugins.sql import flpymssql


LOGGER = logging.get_logger(__name__)


class FLPYODBC(flpymssql.FLPYMSSQL):
    """FLPYODBC class."""

    def __init__(self):
        """Inicialize."""
        super().__init__()
        self.name_ = "FLPYODBC"
        self.alias_ = "SQL Server (PYODBC)"
        self._safe_load = {"pyodbc": "pyodbc", "sqlalchemy": "sqlAlchemy"}
        self._database_not_found_keywords = ["4060"]
        self._sqlalchemy_name = "mssql+pyodbc"

    def loadConnectionString(self, name: str, host: str, port: int, usern: str, passw_: str) -> str:
        """Set special config."""

        return (
            super().loadConnectionString(name, host, port, usern, passw_)
            + "?driver=ODBC+Driver+17+for+SQL+Server"
        )



# -*- coding: utf-8 -*-
"""
Collection of utility functions.

Just an assortment of functions that don't depend on externals and don't fit other modules.
"""


from PyQt6 import QtCore, QtXml, QtWidgets  # type: ignore[import]

from pineboolib.core.utils import logging
from pineboolib.core import settings

import os
import re
import sys
import io
import os.path
import hashlib
import traceback
import types
import threading
import time

from typing import Optional, Union, Any, List, cast, Callable, TypeVar, TYPE_CHECKING

from xml.etree import ElementTree

if TYPE_CHECKING:
    from pineboolib.application.qsatypes.date import Date  # noqa: F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)
T1 = TypeVar("T1")

# FIXME: Move commaSeparator to Pineboo internals, not aqApp
DECIMAL_SEPARATOR = (
    "," if QtCore.QLocale.system().toString(float(0.01), "f", 2).find(",") > -1 else "."
)

BASE_DIR = None
FORCE_DESKTOP = False

SHOWED_FORCE_DESKTOP_WARNING = False


def auto_qt_translate_text(text: Optional[str]) -> str:
    """Remove QT_TRANSLATE from Eneboo XML files. This function does not translate."""
    if not isinstance(text, str):
        text = str(text)

    if isinstance(text, str):
        if text.find("QT_TRANSLATE") != -1:
            match = re.search(r"""QT_TRANSLATE\w*\(.+,["'](.+)["']\)""", text)
            text = match.group(1) if match else ""

    return text


def qt_translate_noop(string: str, path: str, mod: str) -> str:
    """Translate string."""

    if string.find("QT_TRANSLATE_NOOP") == -1:
        return string
    string_list = string[18:-1].split(",")
    string = string_list[1][1:-1]

    nombre_fichero = os.path.join(
        path, "translations", "%s.%s.ts" % (mod, QtCore.QLocale().name()[:2])
    )
    if not os.path.exists(nombre_fichero):
        LOGGER.debug("flreloadlast.traducirCadena: No se encuentra el fichero %s" % nombre_fichero)
        return string

    fichero = open(nombre_fichero, "r", encoding="ISO-8859-15")
    file_data = fichero.read()
    fichero.close()
    xml_translations = QtXml.QDomDocument()
    if xml_translations.setContent(file_data):
        node_mess = xml_translations.elementsByTagName("message")
        for node_number in range(len(node_mess)):
            node = node_mess.item(node_number)
            if node.namedItem("source").toElement().text() == string:
                traduccion = node.namedItem("translation").toElement().text()
                if traduccion:
                    return traduccion

    return string


AQTT = auto_qt_translate_text


def one(list_: List[T1], default: Any = None) -> Optional[T1]:
    """
    Retrieve first element of the list or None/default.

    Useful to avoid try/except cluttering and clean code.
    """
    try:
        return list_[0]
    except IndexError:
        return default


def traceit(
    frame: types.FrameType, event: str, arg: Any  # pylint: disable=unused-argument
) -> Callable[[types.FrameType, str, Any], Any]:
    """
    Print a trace line for each Python line executed or call.

    This function is intended to be the callback of sys.settrace.
    """

    # if event != "line":
    #    return traceit
    try:
        import linecache

        lineno = frame.f_lineno
        filename = frame.f_globals["__file__"]
        # if "pineboo" not in filename:
        #     return traceit
        if filename.endswith(".pyc") or filename.endswith(".pyo"):
            filename = filename[:-1]
        name = frame.f_globals["__name__"]
        line = linecache.getline(filename, lineno)
        print("%s:%s:%s %s" % (name, lineno, event, line.rstrip()))
    except Exception:
        pass
    return traceit


class TraceBlock:
    """
    With Decorator to add traces on a particular block.

    Use it like:

    with TraceBlock():
        code
    """

    def __enter__(self) -> Callable[[types.FrameType, str, Any], Any]:
        """Create tracing context on enter."""
        # NOTE: "sys.systrace" function could lead to arbitrary code execution
        sys.settrace(traceit)  # noqa: DUO111
        return traceit

    def __exit__(self, type: Any, value: Any, traceback: Any) -> None:
        """Remove tracing context on exit."""
        # NOTE: "sys.systrace" function could lead to arbitrary code execution
        sys.settrace(None)  # noqa: DUO111


def trace_function(func_: Callable) -> Callable:
    """Add tracing to decorated function."""

    def wrapper(*args: Any) -> Any:
        with TraceBlock():
            return func_(*args)

    return wrapper


def copy_dir_recursive(from_dir: str, to_dir: str, replace_on_conflict: bool = False) -> bool:
    """
    Copy a folder recursively.

    *** DEPRECATED ***
    Use python shutil.copytree for this.
    """
    dir = QtCore.QDir()
    dir.setPath(from_dir)

    from_dir += QtCore.QDir.separator()
    to_dir += QtCore.QDir.separator()

    if not os.path.exists(to_dir):
        os.makedirs(to_dir)

    for file_ in dir.entryList(QtCore.QDir.Filter.Files):
        from_ = from_dir + file_
        to_ = to_dir + file_
        if str(to_).endswith(".src"):
            to_ = str(to_).replace(".src", "")
            # print("Destino", to_)

        if os.path.exists(to_):
            if replace_on_conflict:
                os.remove(to_)
            else:
                continue

        if not QtCore.QFile.copy(from_, to_):
            return False

    for dir_ in dir.entryList(
        cast(QtCore.QDir.Filter, QtCore.QDir.Filter.Dirs | QtCore.QDir.Filter.NoDotAndDotDot)
    ):
        from_ = from_dir + dir_
        to_ = to_dir + dir_

        if not os.path.exists(to_):
            os.makedirs(to_)

        if not copy_dir_recursive(from_, to_, replace_on_conflict):
            return False

    return True


def text2bool(text: str) -> bool:
    """Convert input text into boolean, if possible."""

    if str(text).lower().startswith(("t", "y", "1", "on", "s")):
        return True
    elif str(text).lower().startswith(("f", "n", "0", "off")):
        return False

    raise ValueError("Valor booleano no comprendido '%s'" % text)


def ustr(*full_text: Union[bytes, str, int, "Date", None, float]) -> str:
    """Convert and concatenate types to text."""

    def ustr1(text_: Union[bytes, str, int, "Date", None, float]) -> str:
        if isinstance(text_, str):
            return text_

        elif isinstance(text_, float):
            t_float = text_
            try:
                text_ = int(text_)

                if not t_float == text_:
                    text_ = str(t_float)

            except Exception:
                pass

            return str(text_)

        elif isinstance(text_, bytes):
            return str(text_, "UTF-8")

        else:
            return repr("" if text_ is None else text_)

    return "".join([ustr1(text) for text in full_text])


class StructMyDict(dict):
    """Dictionary that can be read/written using properties."""

    def __getattr__(self, name: str) -> Any:
        """Get property."""
        try:
            return self[name]
        except KeyError as error:
            raise AttributeError(error)

    def __setattr__(self, name: str, value: Any) -> None:
        """Set property."""
        self[name] = value


def load2xml(form_path_or_str: str) -> ElementTree.ElementTree:
    """Parse a Eneboo style XML."""

    """
    class xml_parser(ET.TreeBuilder):


        def start(self, tag, attrs):
            return super(xml_parser, self).start(tag, attrs)

        def end(self, tag):
            return super(xml_parser, self).end(tag)

        def data(self, data):
            super(xml_parser, self).data(data)

        def close(self):
            return super(xml_parser, self).close()
    """

    file_ptr: Optional[io.StringIO] = None
    if form_path_or_str.find("KugarTemplate") > -1 or form_path_or_str.find("DOCTYPE") > -1:
        form_path_or_str = _parse_for_duplicates(form_path_or_str)
        file_ptr = io.StringIO(form_path_or_str)
    elif not os.path.exists(form_path_or_str):
        raise Exception("File %s not found" % form_path_or_str[:200])

    try:
        parser = ElementTree.XMLParser()
        return ElementTree.parse(file_ptr or form_path_or_str, parser)
    except Exception:
        try:
            parser = ElementTree.XMLParser(encoding="ISO-8859-15")
            return ElementTree.parse(file_ptr or form_path_or_str, parser)
        except Exception:
            """LOGGER.exception(
                "Error cargando UI después de intentar con UTF8 e ISO \n%s", form_path_or_str
            )"""
            raise


def _parse_for_duplicates(text: str) -> str:
    """load2xml helper for Kugar XML."""
    ret_ = ""
    text = text.replace("+", "__PLUS__")
    text = text.replace("-", "__MINUS__")
    text = text.replace("(", "__LPAREN__")
    text = text.replace(")", "__RPAREN__")
    text = text.replace("*", "__ASTERISK__")

    for section_orig in text.split(">"):
        # print("section", section)
        if "<!__MINUS____MINUS__" in section_orig:
            continue

        duplicate_ = False
        attr_list: List[str] = []

        # print("--->", section_orig)
        ret2_ = ""
        section = ""
        for num, action in enumerate(section_orig.split(" ")):
            count_ = action.count("=")
            if count_ > 1:
                # part_ = ""
                text_to_process = action
                for item in range(count_):
                    pos_ini = text_to_process.find('"')

                    pos_fin = text_to_process[pos_ini + 1 :].find('"')
                    # print("Duplicado", item, pos_ini, pos_fin, text_to_process, "***" , text_to_process[0:pos_ini + 2 + pos_fin])
                    ret2_ += " %s " % text_to_process[0 : pos_ini + 2 + pos_fin]
                    text_to_process = text_to_process[pos_ini + 2 + pos_fin :]

            else:
                ret2_ += "%s " % action

        section += ret2_
        if section.endswith(" "):
            section = section[0 : len(section) - 1]

        if section_orig.endswith("/") and not section.endswith("/"):
            section += "/"

        # print("***", section)
        section = section.replace(" =", "=")
        section = section.replace('= "', '="')

        for attribute_ in section.split(" "):
            # print("attribute", attribute_)
            if attribute_.find("=") > -1:
                attr_name = attribute_[0 : attribute_.find("=")]
                if attr_name not in attr_list:
                    attr_list.append(attr_name)
                else:
                    if attr_name != "":
                        # print("Eliminado attributo duplicado", attr_name)
                        duplicate_ = True

            if not duplicate_:
                if not section.endswith(attribute_):
                    ret_ += "%s " % attribute_
                else:
                    ret_ += "%s" % attribute_
            else:
                if attribute_.endswith("/"):
                    ret_ += "/"

            duplicate_ = False

        if (section.find(">") == -1 and section.find("<") > -1) or section.endswith(
            "__MINUS____MINUS__"
        ):
            ret_ += ">"

    # print(ret_)
    ret_ = ret_.replace(">__MINUS____MINUS__", ">")
    ret_ = ret_.replace(" __", "__BLANCK__")
    ret_ = ret_.replace("__ ", "__BLANCK__")
    ret_ = ret_.replace('"__', '" __')
    ret_ = ret_.replace('__"', '__ "')
    # ret_ = ret_.replace('=" __', '"__')
    # ret_ = ret_.replace('__ "', '__"')

    return ret_


def pretty_print_xml(elem: ElementTree.Element, level: int = 0) -> None:
    """
    Generate pretty-printed version of given XML.

    copy and paste from http://effbot.org/zone/element-lib.htm#prettyprint
    it basically walks your tree and adds spaces and newlines so the tree is
    printed in a nice way
    """
    lev_ = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = lev_ + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = lev_
        for elem in elem:
            pretty_print_xml(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = lev_
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = lev_


def format_double(value: Union[int, str, float], part_integer: int, part_decimal: int = 0) -> str:
    """Convert number into string with fixed point style."""
    if isinstance(value, str) and value == "":
        return value

    value_array = str(round(float(str(value)), part_decimal)).split(".")

    str_integer = format_int(value_array[0], part_integer)
    str_decimal = value_array[1] if len(value_array) > 1 and part_decimal else ""

    while part_decimal > len(str_decimal):
        str_decimal += "0"

    # Fixme: Que pasa cuando la parte entera sobrepasa el limite, se coge el maximo valor o
    return "%s%s%s" % (str_integer, DECIMAL_SEPARATOR if str_decimal else "", str_decimal)


def format_int(value: Union[str, int, float, None], part_integer: int = 0) -> str:
    """Convert integer into string."""
    if value is not None:
        str_integer = "{:,d}".format(int(value))
        value = str_integer.replace(DECIMAL_SEPARATOR, "," if DECIMAL_SEPARATOR == "," else ".")
        len_value = len(value)
        return (
            value[len_value - part_integer :]
            if part_integer and part_integer < len_value
            else value
        )
    else:
        return ""


def is_deployed() -> bool:
    """Return wether we're running inside a PyInstaller bundle."""
    return getattr(sys, "frozen", False)


def is_library() -> bool:
    """Return if pineboolib is used as external library."""

    global SHOWED_FORCE_DESKTOP_WARNING

    if FORCE_DESKTOP:
        if not SHOWED_FORCE_DESKTOP_WARNING:
            LOGGER.info("is_library: Force Desktop is Activated!")
            SHOWED_FORCE_DESKTOP_WARNING = True

        return False
    return QtWidgets.QApplication.platformName() == "offscreen"


def get_base_dir() -> str:
    """Obtain pinebolib installation path."""
    global BASE_DIR
    if not BASE_DIR:
        BASE_DIR = "%s/../.." % os.path.dirname(__file__)
        BASE_DIR = (
            os.path.realpath(".%s" % BASE_DIR[1:])
            if is_deployed() and BASE_DIR.startswith(":")
            else BASE_DIR
        )
        LOGGER.info("BaseDir %s", BASE_DIR)
    return BASE_DIR


def filedir(*path: str) -> str:
    """
    Get file full path reltive to the project.

    filedir(path1[, path2, path3 , ...])
    @param array de carpetas de la ruta
    @return devuelve la ruta absoluta resultado de concatenar los paths que se le pasen y aplicarlos desde la ruta del proyecto.
    Es útil para especificar rutas a recursos del programa.
    """
    return os.path.realpath(os.path.join(get_base_dir(), *path))


def download_files() -> None:
    """Download data for PyInstaller bundles."""
    if os.path.exists(filedir("forms")):
        return

    if not os.path.exists(filedir("../pineboolib")):
        os.mkdir(filedir("../pineboolib"))

    copy_dir_recursive(":/pineboolib", filedir("../pineboolib"))

    tmp_dir = settings.CONFIG.value("ebcomportamiento/temp_dir")

    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)


def pixmap_from_mime_source(name: str) -> Any:
    """Convert mime source into a pixmap."""
    from PyQt6 import QtGui

    file_name = filedir("./core/images/icons", name)

    return QtGui.QPixmap(file_name) if os.path.exists(file_name) else None


def sha1(text_: str) -> str:
    """Get SHA1 hash from string in hex form."""
    return hashlib.sha1(str(text_).encode("UTF-8")).hexdigest()


def print_stack(maxsize: int = 1) -> None:
    """Print Python stack, like a traceback."""
    for item in traceback.format_list(traceback.extract_stack())[1:-2][-maxsize:]:
        print(item.rstrip())


def session_id(conn_name: str = "default", with_time: bool = False) -> str:
    """Return session id."""

    return "%s|%s%s" % (
        threading.current_thread().ident,
        conn_name,
        "|%s" % time.time() if with_time else "",
    )


def empty_dir(dir_name: str) -> None:
    """Empty a dir."""

    for root, dirs, files in os.walk(dir_name):
        for file_item in files:
            os.remove(os.path.join(root, file_item))


def sha1(value: Union[str, bytes, None] = "") -> str:
    """Return sha1 value."""

    if value is None:
        value = ""

    if isinstance(value, str):
        value = value.encode()

    sha_ = hashlib.new("sha1", value)
    string = "%s" % sha_.hexdigest()

    return string.upper()



"""
test_pncompoundkeymetadata Module.
"""

import unittest
from pineboolib.loader.main import init_testing


class TestPNCompoundKeyFromTableMetaData(unittest.TestCase):
    """Test string field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test check that the data is correct in a compound key."""
        from pineboolib import application

        mtd = application.PROJECT.conn_manager.manager().metadata("flseqs")
        if mtd is None:
            raise Exception
        field_1 = mtd.field("campo")
        if field_1 is None:
            raise Exception

        self.assertEqual(field_1.name(), "campo")
        self.assertEqual(field_1.isCompoundKey(), True)

        ck_list = mtd.fieldListOfCompoundKey("campo")
        if ck_list is None:
            raise Exception
        self.assertEqual(len(ck_list), 1)
        self.assertEqual(ck_list[0].name(), "campo")


class TestGeneratePNCompoundKey(unittest.TestCase):
    """Test string field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test create a composite key and check its values."""
        from pineboolib import application
        from pineboolib.application.metadata.pncompoundkeymetadata import PNCompoundKeyMetaData

        mtd = application.PROJECT.conn_manager.manager().metadata("flfiles")
        if mtd is None:
            raise Exception

        field_1 = mtd.field("sha")
        field_2 = mtd.field("contenido")
        if field_1 is None:
            raise Exception
        if field_2 is None:
            raise Exception
        ck_mtd_1 = PNCompoundKeyMetaData()
        ck_mtd_1.addFieldMD(field_1)
        ck_mtd_1.addFieldMD(field_2)
        field_list_1 = ck_mtd_1.fieldList()

        ck_mtd_2 = PNCompoundKeyMetaData(ck_mtd_1)  # Copy

        field_list_2 = ck_mtd_2.fieldList()

        self.assertEqual(len(field_list_1), 2)
        self.assertEqual(len(field_list_2), 2)
        self.assertEqual(ck_mtd_2.hasField("sha"), True)
        self.assertEqual(ck_mtd_2.hasField("other"), False)


if __name__ == "__main__":
    unittest.main()
